//=============================================================================
//
//	Copyright (C) 1995,1996,1997,1998 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableBorderIO.cc
//
//		Read/write routines for the MiscTableBorder class.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableBorderIO.cc,v 1.4 98/03/24 04:51:24 sunshine Exp $
// $Log:		MiscTableBorderIO.cc,v $
//	Revision 1.4  98/03/24	04:51:24  sunshine
//	v138: Declared 'i' multiple times in MiscTableBorder::write().
//	
//	Revision 1.3  97/12/19	00:55:34  zarnuk
//	v133: Combined sort-dir and sort-type.	Eliminated data-sizing entirely.
//	Eliminated constrain-max functionality.	 Added min/max uniform size.
//	Now passes the version to read().  New storage layout.
//-----------------------------------------------------------------------------
#include "MiscTableBorder.h"
#include "MiscTableScrollPrivate.h"
#include "MiscTableUtil.h"

//-----------------------------------------------------------------------------
// MiscTableSlot::write
//-----------------------------------------------------------------------------
void MiscTableSlot::write( NXTypedStream* stream )
	{
	NXWriteType( stream, @encode(MiscPixels), &size );
	NXWriteType( stream, @encode(MiscPixels), &min_size );
	NXWriteType( stream, @encode(MiscPixels), &max_size );
	NXWriteType( stream, @encode(MiscTableSizing), &sizing );
	}


//-----------------------------------------------------------------------------
// write
//-----------------------------------------------------------------------------
void MiscTableBorder::write( NXTypedStream* stream )
	{
	register unsigned int m = 0;
	m = (m << 1) | (prototypes != 0);
	m = (m << 1) | (titles != 0);
	m = (m << 1) | (styles != 0);
	m = (m << 1) | (tags != 0);
	m = (m << 1) | (sort_info != 0);
	m = (m << 1) | (p2v != 0);
	m = (m << 1) | (slots != 0);
	m = (m << 1) | (modifier_drag != 0);
	m = (m << 1) | (draggable != 0);
	m = (m << 1) | (sizeable != 0);
	m = (m << 1) | (selectable != 0);
	m = (m << 1) | (type != 0);

	unsigned int const mbuff = m;
	NXWriteType( stream, @encode(unsigned int), &mbuff );
	NXWriteObjectReference( stream, owner );
	def_slot.write( stream );
	NXWriteType( stream, @encode(MiscTableCellStyle), &def_style );
	NXWriteType( stream, @encode(int), &def_tag );
	NXWriteType( stream, @encode(MiscTableTitleMode), &title_mode );
	NXWriteType( stream, @encode(MiscPixels), &min_uniform_size );
	NXWriteType( stream, @encode(MiscPixels), &max_uniform_size );
	NXWriteType( stream, @encode(MiscPixels), &uniform_size );
	NXWriteType( stream, @encode(MiscPixels), &min_total_size );
	NXWriteType( stream, @encode(int), &num_springy );
	NXWriteType( stream, @encode(int), &num_slots );

	int i;
	if (slots != 0)
		for (i = 0; i < num_slots; i++)
			slots[i].write( stream );

	if (p2v != 0)
		NXWriteArray( stream, @encode(MiscCoord_V), num_slots, p2v );

	if (sort_info != 0)
		NXWriteArray( stream, @encode(int), num_slots, sort_info );

	if (tags != 0)
		NXWriteArray( stream, @encode(int), num_slots, tags );

	if (styles != 0)
		NXWriteArray( stream, @encode(MiscTableCellStyle), num_slots, styles );

	if (titles != 0)
		NXWriteArray( stream, @encode(char*), num_slots, titles );

	if (prototypes != 0)
		for (i = 0; i < num_slots; i++)
			NXWriteObject( stream, prototypes[i] );
	}


//-----------------------------------------------------------------------------
// read
//-----------------------------------------------------------------------------
void MiscTableBorder::read( NXTypedStream* stream, int ver )
	{
	if (ver < MISC_TS_VERSION_2)
		read_v1( stream, ver );
	else
		read_v2( stream, ver );
	}


//-----------------------------------------------------------------------------
// MiscTableSlot::read
//-----------------------------------------------------------------------------
void MiscTableSlot::read( NXTypedStream* stream, int version )
	{
	NXReadType( stream, @encode(MiscPixels), &size );
	NXReadType( stream, @encode(MiscPixels), &min_size );
	NXReadType( stream, @encode(MiscPixels), &max_size );
	if (version < MISC_TS_VERSION_2)
		{
		MiscPixels data_size;
		unsigned int x;
		NXReadType( stream, @encode(MiscPixels), &data_size );
		NXReadType( stream, @encode(MiscTableSizing), &x );
		sizing = MiscTableSizing( (x & 1) | ((x & 4) >> 1) );
		}
	else
		NXReadType( stream, @encode(MiscTableSizing), &sizing );
	MISC_ENUM_CHECK( sizing, MISC_MAX_SIZING );
	offset = 0;
	adj_size = size;
	}


//-----------------------------------------------------------------------------
// read_v2
//-----------------------------------------------------------------------------
void MiscTableBorder::read_v2( NXTypedStream* stream, int ver )
	{
	emptyAndFree();

	sort_funcs = 0;		// Can't archive function addresses.

	unsigned int mbuff;
	NXReadType( stream, @encode(unsigned int), &mbuff );
	register unsigned int m = mbuff;

	type		= MiscBorderType( m & 1 );
	selectable	= bool((m >>= 1) & 1);
	sizeable	= bool((m >>= 1) & 1);
	draggable	= bool((m >>= 1) & 1);
	modifier_drag = bool((m >>= 1) & 1);

	owner = NXReadObject( stream );
	def_slot.read( stream, ver );
	NXReadType( stream, @encode(MiscTableCellStyle), &def_style );
	NXReadType( stream, @encode(int), &def_tag );
	NXReadType( stream, @encode(MiscTableTitleMode), &title_mode );
	NXReadType( stream, @encode(MiscPixels), &min_uniform_size );
	NXReadType( stream, @encode(MiscPixels), &max_uniform_size );
	NXReadType( stream, @encode(MiscPixels), &uniform_size );
	NXReadType( stream, @encode(MiscPixels), &min_total_size );
	NXReadType( stream, @encode(int), &num_springy );
	NXReadType( stream, @encode(int), &num_slots );
	max_slots = num_slots;

	if ((m >>= 1) & 1)
		{
		alloc_slots();
		for (int i = 0; i < num_slots; i++)
			slots[i].read( stream, ver );
		}

	if ((m >>= 1) & 1)
		{
		alloc_vmap();
		NXReadArray( stream, @encode(MiscCoord_V), num_slots, p2v );
		for (int i = 0; i < num_slots; i++)
			v2p[ p2v[i] ] = i;
		}

	if ((m >>= 1) & 1)
		{
		alloc_sort_info();
		NXReadArray( stream, @encode(int), num_slots, sort_info );
		}

	if ((m >>= 1) & 1)
		{
		alloc_tags();
		NXReadArray( stream, @encode(int), num_slots, tags );
		}

	if ((m >>= 1) & 1)
		{
		alloc_styles();
		NXReadArray( stream, @encode(MiscTableCellStyle), num_slots, styles );
		}

	if ((m >>= 1) & 1)
		{
		alloc_titles();
		NXReadArray( stream, @encode(char*), num_slots, titles );
		}

	if ((m >>= 1) & 1)
		{
		alloc_prototypes();
		for (int i = 0;	 i < num_slots;	 i++)
			prototypes[i] = NXReadObject( stream );
		}

	needs_recalc = true;
	}


//-----------------------------------------------------------------------------
// read_v1
//-----------------------------------------------------------------------------
void MiscTableBorder::read_v1( NXTypedStream* stream, int ver )
	{
	int n;
	emptyAndFree();

	NXReadType( stream, @encode(MiscBorderType), &type );
	MISC_ENUM_CHECK( type, MISC_MAX_BORDER );
	owner = NXReadObject( stream );
	def_slot.read( stream, ver );
	NXReadType( stream, @encode(int), &num_slots );
	max_slots = num_slots;

	NXReadType( stream, @encode(int), &n );
	if (n != 0)
		{
		alloc_slots();
		for (int i = 0;	 i < num_slots;	 i++)
			slots[i].read( stream, ver );
		}

	NXReadType( stream, @encode(int), &n );
	if (n != 0)
		{
		alloc_vmap();
		for (int i = 0;	 i < num_slots;	 i++)
			NXReadType( stream, @encode(MiscCoord_P), &(v2p[i]) );
		for (int j = 0;	 j < num_slots;	 j++)
			NXReadType( stream, @encode(MiscCoord_V), &(p2v[j]) );
		}

	NXReadType( stream, @encode(int), &def_tag );
	NXReadType( stream, @encode(int), &n );
	if (n != 0)
		{
		alloc_tags();
		for (int i = 0;	 i < num_slots;	 i++)
			{
			int const j = visualToPhysical(i);
			NXReadType( stream, @encode(int), &(tags[j]) );
			}
		}

	NXReadType( stream, @encode(MiscPixels), &uniform_size );
	NXReadType( stream, @encode(MiscPixels), &min_total_size );
	MiscPixels max_total_size;	// obsolete.
	NXReadType( stream, @encode(MiscPixels), &max_total_size );
	NXReadType( stream, @encode(int), &num_springy );
	NXReadType( stream, @encode(MiscTableTitleMode), &title_mode );
	MISC_ENUM_CHECK( title_mode, MISC_MAX_TITLE );

	NXReadType( stream, @encode(int), &n );
	if (n != 0)
		{
		alloc_titles();
		for (int i = 0;	 i < num_slots;	 i++)
			{
			int const j = visualToPhysical(i);
			NXReadType( stream, @encode(char*), &(titles[j]) );
			}
		}

	NXReadType( stream, @encode(MiscTableCellStyle), &def_style );
	MISC_ENUM_CHECK( def_style, MISC_TABLE_CELL_MAX );

	NXReadType( stream, @encode(int), &n );
	if (n != 0)
		{
		alloc_styles();
		for (int i = 0;	 i < num_slots;	 i++)
			{
			int const j = visualToPhysical(i);
			NXReadType( stream, @encode(MiscTableCellStyle), &(styles[j]) );
			MISC_ENUM_CHECK( styles[j], MISC_TABLE_CELL_MAX );
			}
		}

	NXReadType( stream, @encode(int), &n );
	if (n != 0)
		{
		alloc_prototypes();
		for (int i = 0;	 i < num_slots;	 i++)
			{
			int const j = visualToPhysical(i);
			prototypes[j] = NXReadObject(stream);
			}
		}

	sort_funcs = 0;		// Can't archive function addresses.

	NXReadType( stream, @encode(int), &n );
	if (n != 0)
		{
		alloc_sort_info();
		for (int i = 0;	 i < num_slots;	 i++)
			{
			int const j = visualToPhysical(i);
			MiscSortDirection x;
			NXReadType( stream, @encode(MiscSortDirection), &x );
			MISC_ENUM_CHECK( x, MISC_SORT_DIR_MAX );
			int const z = sort_info[j];
			sort_info[j] = ((z & ~1) | (x & 1));
			}
		}

	NXReadType( stream, @encode(int), &n );
	if (n != 0)
		{
		if (sort_info == 0)
			alloc_sort_info();
		for (int i = 0;	 i < num_slots;	 i++)
			{
			int const j = visualToPhysical(i);
			MiscSortType x;
			NXReadType( stream, @encode(MiscSortType), &x );
			MISC_ENUM_CHECK( x, MISC_SORT_TYPE_MAX );
			int const z = sort_info[j];
			sort_info[j] = ((x << 1) | (z & 1));
			}
		}

	NXReadType( stream, @encode(bool), &selectable );
	NXReadType( stream, @encode(bool), &sizeable );
	NXReadType( stream, @encode(bool), &draggable );
	NXReadType( stream, @encode(bool), &modifier_drag );

	min_uniform_size = MISC_MIN_PIXELS_SIZE;
	max_uniform_size = MISC_MAX_PIXELS_SIZE;
	needs_recalc = true;
	}
