//=============================================================================
//
//		Copyright (C) 1995-1997 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableBorderSel.cc
//
//		Selection management methods for the MiscTableBorder class.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableBorderSel.cc,v 1.2 97/11/16 17:54:35 zarnuk Exp $
// $Log:		MiscTableBorderSel.cc,v $
//	Revision 1.2  97/11/16	17:54:35  zarnuk
//	v132: Renamed.
//	
//	Revision 1.1  97/11/14	20:34:41  zarnuk
//	v130: selection management routines for MiscTableBorder.
//	
//-----------------------------------------------------------------------------
#include "MiscTableBorder.h"
#include <misckit/MiscIntList.h>


//-----------------------------------------------------------------------------
// visualToPhysical
//		Used to save the selection before sorting.
//-----------------------------------------------------------------------------
void MiscTableBorder::visualToPhysical( MiscSparseSet const& vset,
										MiscSparseSet& pset ) const
	{
	pset.empty();
	unsigned int const lim = vset.numRanges();
	for (unsigned int i = 0; i < lim; i++)
		{
		int lo, hi;
		vset.getRangeAt( i, lo, hi );
		for ( ; lo <= hi; lo++)
			pset.add( visualToPhysical( lo ) );
		}
	}


//-----------------------------------------------------------------------------
// physicalToVisual
//		Used to restore the selection after sorting.
//-----------------------------------------------------------------------------
void MiscTableBorder::physicalToVisual( MiscSparseSet const& pset,
										MiscSparseSet& vset ) const
	{
	vset.empty();
	unsigned int const lim = pset.numRanges();
	for (unsigned int i = 0; i < lim; i++)
		{
		int lo, hi;
		pset.getRangeAt( i, lo, hi );
		for ( ; lo <= hi; lo++)
			vset.add( physicalToVisual( lo ) );
		}
	}


//-----------------------------------------------------------------------------
// hasMultipleSelection
//-----------------------------------------------------------------------------
bool MiscTableBorder::hasMultipleSelection() const
	{
	MiscCoord_V lo, hi;
	selection.getTotalRange( lo, hi );
	return hi > lo;
	}


//-----------------------------------------------------------------------------
// selected_slots
//-----------------------------------------------------------------------------
void MiscTableBorder::selected_slots( MiscIntList* list, bool do_tags ) const
	{
	[list empty];
	for (unsigned int i = 0, lim = selection.numRanges(); i < lim; i++)
		{
		MiscCoord_V lo, hi;
		selection.getRangeAt( i, lo, hi );
		for ( ; lo <= hi; lo++)
			[list addInt: (do_tags ? getTag(lo) : visualToPhysical(lo))];
		}
	}


//-----------------------------------------------------------------------------
// select_slots
//-----------------------------------------------------------------------------
void MiscTableBorder::select_slots( MiscIntList* list, bool clear, bool set )
	{
	if (clear)
		selectNone();
	int const lim = [list count];
	if (lim > 0)
		{
		MiscCoord_V last_selected = -1;
		for (int i = lim; i-- > 0; )
			{
			MiscCoord_P const p_slot = [list intAt:i];
			if (goodPos( p_slot ))
				{
				MiscCoord_V const v_slot = physicalToVisual( p_slot );
				if (v_slot > last_selected) last_selected = v_slot;
				if (set)
					selection.add( v_slot );
				else
					selection.remove( v_slot );
				}
			}
		if (set && last_selected != -1)
			setSelectedSlot( last_selected );
		else
			fixSelectedSlot();
		}
	}


//-----------------------------------------------------------------------------
// select_tags
//-----------------------------------------------------------------------------
void MiscTableBorder::select_tags( MiscIntList* list, bool clear, bool set )
	{
	if (clear)
		selectNone();
	unsigned int const M = [list count];
	if (M > 0)
		{
		MiscCoord_V last_selected = -1;
		int const* const v0 = [list rawData];
		int const* const vM = v0 + M;

		unsigned int const N = count();
		for (unsigned int i = 0; i < N; i++)
			{
			int const t = getTag(i);			// MiscCoord_V
			for (int const* v = v0; v < vM; v++)
				if (*v == t)
					{
					if (i > last_selected) last_selected = (MiscCoord_V)i;
					if (set)
						selection.add(i);
					else
						selection.remove(i);
					break;
					}
			}
		if (set && last_selected != -1)
			setSelectedSlot( last_selected );
		else
			fixSelectedSlot();
		}
	}
