//=============================================================================
//
//	Copyright (C) 1995,1996,1997,1998 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableScroll.M
//
//		ScrollView class that displays a 2-D table of cells.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableScroll.M,v 1.39 98/03/24 04:50:22 sunshine Exp $
// $Log:		MiscTableScroll.M,v $
// Revision 1.39  98/03/24	04:50:22  sunshine
// v138: Removed unused/junk -calcWidths:heights:maxW:maxH:.
// Removed useless 'if' from -borderSetSlotOrder:.
// Fixed bug: Wasn't checking -canDraw before applying -lockFocus in:
// -drawCellAt::, -drawRow:, -drawCol:, -drawSlotTitle:.
// 
// Revision 1.38  98/02/25	13:46:45  zarnuk
// v137: border:setSlotOrder: now accepts a null pointer for
// the list argument to "unsort" the slots.
// 
// Revision 1.37  97/12/22	21:58:09  zarnuk
// v134: now uses suspendEditing / resumeEditing.
//-----------------------------------------------------------------------------
#import <misckit/MiscTableScroll.h>
#import <misckit/MiscTableCell.h>
#import <misckit/MiscIntList.h>
#import "MiscColView.h"
#import "MiscCornerView.h"
#import "MiscRowView.h"
#import "MiscTableBorder.h"
#import "MiscTableScrollPrivate.h"
#import "MiscTableView.h"
#import <new.h>

extern "Objective-C" {
#import <appkit/Application.h>
#import <appkit/Cell.h>
#import <appkit/ClipView.h>
#import <appkit/Font.h>
#import <appkit/FontManager.h>
#import <appkit/Scroller.h>
}

extern "C" {
#import <math.h>
#import <strings.h>
}

typedef MiscDelegateFlags DF;

@implementation MiscTableScroll

- (int)tag				{ return tag; }
- (id)setTag:(int)x		{ tag = x;	return self; }

//-----------------------------------------------------------------------------
// + initialize
//-----------------------------------------------------------------------------
+ (id)initialize
	{
	if (self == [MiscTableScroll class])
		[self setVersion: MISC_TS_VERSION];
	return self;
	}


//-----------------------------------------------------------------------------
// Multicast Messages
//-----------------------------------------------------------------------------
- (id)sendAction:(SEL)aSel to:(id)obj forAllCells:(BOOL)flag
		{
		int const rlim = num_rows;
		int const clim = num_cols;
		for (int r = 0;	 r < rlim;	r++)
			for (int c = 0;	 c < clim;	c++)
				if (flag || [self cellIsSelected:r:c])
					if (![obj perform:aSel with:[self cellAt:r:c]])
						return self;
		return self;
		}

- (int)makeCellsPerform:(SEL)s with:(id)p1 with:(id)p2 selectedOnly:(BOOL)f
		{
		int count = 0;
		int const rlim = num_rows;
		int const clim = num_cols;
		for (int r = 0;	 r < rlim;	r++)
			for (int c = 0;	 c < clim;	c++)
				if (!f || [self cellIsSelected:r:c])
					{
					id cell = [self cellAt:r:c];
					if ([cell respondsTo:s])
						if ([cell perform:s with:p1 with:p2])
							count++;
						else
							return count;
					}
		return count;
		}

- (int)makeCellsPerform:(SEL)aSel with:(id)p1 selectedOnly:(BOOL)flag
	{ return [self makeCellsPerform:aSel with:p1 with:0 selectedOnly:flag]; }
- (int)makeCellsPerform:(SEL)aSel selectedOnly:(BOOL)flag
	{ return [self makeCellsPerform:aSel with:0 with:0 selectedOnly:flag]; }

- (int)makeCellsPerform:(SEL)aSel
	{ return [self makeCellsPerform:aSel selectedOnly:NO]; }
- (int)makeCellsPerform:(SEL)aSel with:(id)p1
	{ return [self makeCellsPerform:aSel with:p1 selectedOnly:NO]; }
- (int)makeCellsPerform:(SEL)aSel with:(id)p1 with:(id)p2
	{ return [self makeCellsPerform:aSel with:p1 with:p2 selectedOnly:NO];}


//-----------------------------------------------------------------------------
// FINDING CELLS
//-----------------------------------------------------------------------------
- (int)border:(MiscBorderType)b findSlotWithTag:(int)x
		{
		int const lim = (int) [self numSlots:b];
		for (int i = 0;	 i < lim;  i++)
			if ([self border:b slotTag:i] == x)
				return i;
		return -1;
		}
- (int)findColWithTag:(int)x
		{ return [self border:MISC_COL_BORDER findSlotWithTag:x]; }
- (int)findRowWithTag:(int)x
		{ return [self border:MISC_ROW_BORDER findSlotWithTag:x]; }

- (id)findCell:(id)cell row:(int*)row col:(int*)col
	{
	int const NRows = [self numRows];
	int const NCols = [self numCols];
	for (int r = 0;	 r < NRows;	 r++)
		for (int c = 0;	 c < NCols;	 c++)
			if ([self cellAt:r:c] == cell)
				{
				*row = r;
				*col = c;
				return self;
				}
	*row = -1;
	*col = -1;
	return 0;
	}

- (id)findCellWithTag:(int)x row:(int*)row col:(int*)col
	{
	int const NRows = [self numRows];
	int const NCols = [self numCols];
	for (int r = 0;	 r < NRows;	 r++)
		for (int c = 0;	 c < NCols;	 c++)
			{
			id const cell = [self cellAt:r:c];
			if (cell && [cell respondsTo:@selector(tag)] && [cell tag] == x)
				{
				*row = r;
				*col = c;
				return cell;
				}
			}
	*row = -1;
	*col = -1;
	return 0;
	}

- (id)findCellWithTag:(int)x
	{
	int r, c;
	return [self findCellWithTag:x row:&r col:&c];
	}



//-----------------------------------------------------------------------------
// - getDocClipFrame:
//-----------------------------------------------------------------------------
- (void)getDocClipFrame:(NXRect*)rect { [contentView getFrame:rect]; }


//-----------------------------------------------------------------------------
// - freeBorder:
//-----------------------------------------------------------------------------
- (void)freeBorder:(MiscBorderInfo*)p
	{
	[p->view removeFromSuperview];
	[p->view free];
	[p->clip removeFromSuperview];
	[p->clip free];
	delete p->border;
	}


//-----------------------------------------------------------------------------
// - free
//-----------------------------------------------------------------------------
- (id)free
	{
	[self setAutodisplay:NO];			// Don't do any drawing!
	[self emptyAndFreeCells];
	[self freeBorder:&colInfo];
	[self freeBorder:&rowInfo];
	[cornerView removeFromSuperview];
	[cornerView free];
	[tableView removeFromSuperview];
	[tableView free];
	delete delegateFlags;
	delete dataDelegateFlags;
	return [super free];
	}


//-----------------------------------------------------------------------------
// - totalSize / totalWidth / totalHeight
//-----------------------------------------------------------------------------
- (NXCoord)totalSize:(MiscBorderType)b
		{ return (NXCoord) info[b]->border->totalSize(); }
- (NXCoord)totalWidth
		{ return [self totalSize:MISC_COL_BORDER]; }
- (NXCoord)totalHeight
		{ return [self totalSize:MISC_ROW_BORDER]; }


//-----------------------------------------------------------------------------
// - constrainSize
//-----------------------------------------------------------------------------
- (void)constrainSize
	{
	MiscTableBorder* b;
	NXRect r;
	[self getDocClipFrame:&r];

	b = colInfo.border;
	if (b->numSpringy() != 0)
		b->setMinTotalSize( (MiscPixels) r.size.width );

	b = rowInfo.border;
	if (b->numSpringy() != 0)
		b->setMinTotalSize( (MiscPixels) r.size.height );

	[rowInfo.view adjustSize];
	[colInfo.view adjustSize];
	[tableView adjustSize];
	}


//-----------------------------------------------------------------------------
// - sizeToCells
//-----------------------------------------------------------------------------
- (id)sizeToCells
	{
	[self constrainSize];
	return self;
	}


//-----------------------------------------------------------------------------
// set_sizes
//-----------------------------------------------------------------------------
static void set_sizes( MiscTableBorder* b, NXCoord const* v, NXCoord lim )
	{
	int const n = b->count();
	if (!b->isUniformSize())
		{
		for (int i = 0; i < n; i++)
			{
			MiscPixels x = (MiscPixels) v[i];
			MiscPixels const xmin = b->getMinSize_P(i);
			MiscPixels const xmax = b->getMaxSize_P(i);
			if (x < xmin) x = xmin;
			if (x > xmax) x = xmax;
			b->setSize_P( i, (MiscPixels) x );
			}
		}
	else if (lim != 0)
		b->setUniformSize( (MiscPixels) lim );
	}


//-----------------------------------------------------------------------------
// - sizeToFit
//-----------------------------------------------------------------------------
- (id)sizeToFit
	{
	NXCoord* const vw = (NXCoord*)calloc( (num_rows + num_cols), sizeof(*vw) );
	NXCoord* const vh = vw + num_cols;
	NXCoord max_h = 0;
	NXCoord max_w = 0;
	for (int r = 0; r < num_rows; r++)
		{
		NXCoord h = 0;
		for (int c = 0; c < num_cols; c++)
			{
			id const cell = [self cellAt:r:c];
			if (cell != 0 && [cell respondsTo:@selector(calcCellSize:)])
				{
				NXSize sz;
				[cell calcCellSize:&sz];
				if (vw[c] < sz.width) vw[c] = sz.width;
				if (max_w < sz.width) max_w = sz.width;
				if (h < sz.height) h = sz.height;
				}
			}
		if (vh[r] < h) vh[r] = h;
		if (max_h < h) max_h = h;
		}

	set_sizes( colInfo.border, vw, max_w );
	set_sizes( rowInfo.border, vh, max_h );
	free( vw );

	return [self sizeToCells];
	}


//-----------------------------------------------------------------------------
// - sizeTo::
//-----------------------------------------------------------------------------
- (id)sizeTo:(NXCoord)width :(NXCoord)height
	{
	[super sizeTo:width:height];
	[self constrainSize];
	return self;
	}


//-----------------------------------------------------------------------------
// - forwardBGColor
//-----------------------------------------------------------------------------
- (void)forwardBGColor
	{
	[super setBackgroundColor:backgroundColor];
	[colInfo.clip setBackgroundColor:backgroundColor];
	[rowInfo.clip setBackgroundColor:backgroundColor];
	}


//-----------------------------------------------------------------------------
// - initBorder:type:
//-----------------------------------------------------------------------------
- (void)initBorder:(MiscBorderInfo*)p type:(MiscBorderType)type
	{
	NXRect emptyRect = {{0,0},{0,0}};
	NXZone* const z = [self zone];
	if (p->border == 0)
		p->border = new( NXZoneMalloc(z,sizeof(*(p->border))) )
						MiscTableBorder( type );
	p->border->setOwner( self );

	if (type == MISC_COL_BORDER)
		p->view = [[MiscColView allocFromZone:z] 
						initFrame:&emptyRect scroll:self info:p->border];
	else
		p->view = [[MiscRowView allocFromZone:z]
						initFrame:&emptyRect scroll:self info:p->border];

	p->clip = [[ClipView allocFromZone:z] initFrame:0];
	[p->clip setDocView:p->view];
	if (p->isOn)
		{
		[self addSubview: p->clip];
		[window invalidateCursorRectsForView: p->view];
		}
	}


//-----------------------------------------------------------------------------
// - doInit:cornerTitle:
//-----------------------------------------------------------------------------
- (void)doInit:(int)ver cornerTitle:(char const*)s
	{
	editInfo.editing = NO;
	tracking = NO;

	[window disableDisplay];
	[self setBorderType: NX_BEZEL];
	[super setHorizScrollerRequired: YES];
	[super setVertScrollerRequired: YES];

	info[ MISC_COL_BORDER ] = &colInfo;
	info[ MISC_ROW_BORDER ] = &rowInfo;

	[self initBorder:&colInfo type:MISC_COL_BORDER];
	[self initBorder:&rowInfo type:MISC_ROW_BORDER];

	NXZone* const z = [self zone];

	delegateFlags =
		new( NXZoneMalloc(z,sizeof(*delegateFlags)) ) MiscDelegateFlags;
	dataDelegateFlags =
		new( NXZoneMalloc(z,sizeof(*dataDelegateFlags)) ) MiscDelegateFlags;

	NXRect emptyRect = {{0,0},{0,0}};
	tableView = [[MiscTableView allocFromZone:z] initFrame:&emptyRect
		scroll:self
		colInfo:colInfo.border
		rowInfo:rowInfo.border];

	cornerView = [[MiscCornerView allocFromZone:z] initFrame:&emptyRect];
	[self setCornerTitle:s];

	if (colInfo.isOn && rowInfo.isOn)
		[self addSubview:cornerView];

	id oldView = [self setDocView: tableView];
	[oldView free];
	[self tile];
	[self constrainSize];
	[self forwardBGColor];

	[self registerServicesTypes];
	[window reenableDisplay];
	}


//-----------------------------------------------------------------------------
// - initFrame:
//-----------------------------------------------------------------------------
- (id)initFrame:(NXRect const*)frameRect
	{
	[super initFrame: frameRect];

	tag = 0;
	enabled = YES;

	colInfo.border = 0;
	colInfo.isOn = YES;

	rowInfo.border = 0;
	rowInfo.isOn = NO;

	id classObj = [self class];
	font = [classObj defaultFont];
	textColor = [classObj defaultTextColor];
	backgroundColor = [classObj defaultBackgroundColor];
	highlightTextColor = [classObj defaultHighlightTextColor];
	highlightBackgroundColor = [classObj defaultHighlightBackgroundColor];

	[self doInit:MISC_TS_VERSION cornerTitle:0];
	return self;
	}


//-----------------------------------------------------------------------------
// - setFrame:
//-----------------------------------------------------------------------------
- (id)setFrame:(NXRect const*)frameRect
	{
	[super setFrame:frameRect];
	[self constrainSize];
	return self;
	}


//-----------------------------------------------------------------------------
// - tile
//-----------------------------------------------------------------------------
- (id)tile
	{
	[window disableDisplay];

	static View* dummy = 0;
	if (dummy == 0)
		{
		NXRect const r = {{0,0},{1,1}};
		dummy = [[View alloc] initFrame:&r];
		}

	View* const old = contentView;
	contentView = dummy;
	[super tile];
	contentView = old;

	NXRect docRect;
	[dummy getFrame:&docRect];

	if (colInfo.isOn != rowInfo.isOn)			// One on, one off.
		{
		MiscBorderInfo& b = (colInfo.isOn ? colInfo : rowInfo);
		int const edge = (colInfo.isOn ? NX_YMIN : NX_XMIN);
		NXCoord height = [b.view frameHeight];
		NXRect rect;

		NXDivideRect( &docRect, &rect, height, edge );

		[b.clip setFrame: &rect];
		[contentView setFrame: &docRect];
		[window invalidateCursorRectsForView: b.view];
		}
	else if (colInfo.isOn && rowInfo.isOn)		// Both on.
		{
		NXCoord colHeight = [colInfo.view frameHeight];
		NXCoord rowWidth  = [rowInfo.view frameHeight];
		NXRect gapRect;
		NXRect colRect;
		NXRect rowRect;

		NXDivideRect( &docRect, &colRect, colHeight, NX_YMIN );
		NXDivideRect( &colRect, &gapRect, rowWidth, NX_XMIN );
		NXDivideRect( &docRect, &rowRect, rowWidth, NX_XMIN );

		[cornerView setFrame: &gapRect];
		[colInfo.clip setFrame: &colRect];
		[rowInfo.clip setFrame: &rowRect];
		[contentView setFrame: &docRect];
		[window invalidateCursorRectsForView: colInfo.view];
		[window invalidateCursorRectsForView: rowInfo.view];
		}
	else										// Both off.
		{
		[contentView setFrame: &docRect];
		}

	[window reenableDisplay];
	return self;
	}


//-----------------------------------------------------------------------------
// - reflectScroll:
//-----------------------------------------------------------------------------
- (id)reflectScroll:(id)aView
	{
	if (aView == contentView) // only reflect position of contentView
		[super reflectScroll: aView];
	return self;
	}


//-----------------------------------------------------------------------------
// - scrollClip:to:
//-----------------------------------------------------------------------------
- (id)scrollClip:(id)aClipView to:(NXPoint const*)aPoint
	{
	if (aClipView == contentView)				// contentView only.
		{
		NXRect rect;
		[window disableFlushWindow];

		[aClipView rawScroll: aPoint];			// Scroll content

		if (colInfo.isOn)
			{
			[colInfo.clip getBounds: &rect];	// Maybe scroll col headings.
			if (rect.origin.x != aPoint->x)
				{
				rect.origin.x = aPoint->x;
				[colInfo.clip rawScroll: &rect.origin];
				[window invalidateCursorRectsForView: colInfo.view];
				}
			}

		if (rowInfo.isOn)
			{
			[rowInfo.clip getBounds: &rect];	// Maybe scroll row labels.
			if (rect.origin.y != aPoint->y)
				{
				rect.origin.y = aPoint->y;
				[rowInfo.clip rawScroll: &rect.origin];
				[window invalidateCursorRectsForView: rowInfo.view];
				}
			}

		[[window reenableFlushWindow] flushWindow];
		}
	else
		{
		BOOL ok = YES;
		NXRect rect;
		[contentView getBounds: &rect];
		if (aClipView == (id)colInfo.clip)
			rect.origin.x = aPoint->x;
		else if (aClipView == (id)rowInfo.clip)
			rect.origin.y = aPoint->y;
		else
			ok = NO;
		if (ok)
			{
			[self scrollClip:contentView to:&(rect.origin)];
			[self reflectScroll:contentView];
			}
		}
	return self;
	}



//-----------------------------------------------------------------------------
// Delegate Stuff
//-----------------------------------------------------------------------------
- (id)delegate			{ return delegate; }
- (id)dataDelegate		{ return dataDelegate; }
- (id)setDelegate:(id)obj
	{ delegate = obj; delegateFlags->setDelegate( obj ); return self; }
- (id)setDataDelegate:(id)obj
	{ dataDelegate = obj; dataDelegateFlags->setDelegate( obj ); return self; }

- (id)responsibleDelegate:(MiscDelegateFlags::Selector)cmd
	{
	id del = 0;
	if (delegate != 0 && delegateFlags->respondsTo(cmd))
		del = delegate;
	else if (dataDelegate != 0 && dataDelegateFlags->respondsTo(cmd))
		del = dataDelegate;
	return del;
	}

- (char const*)border:(MiscBorderType)b getDelegateSlotTitle:(int)slot
	{
	id del = [self responsibleDelegate:DF::DEL_SLOT_TITLE];
	if (del != 0)
		return [del tableScroll:self border:b slotTitle:slot];
	return 0;
	}

- (id)border:(MiscBorderType)b getDelegateSlotPrototype:(int)s
	{
	id del = [self responsibleDelegate:DF::DEL_SLOT_PROTOTYPE];
	if (del != 0)
		return [del tableScroll:self border:b slotPrototype:s];
	return 0;
	}

- (void)border:(MiscBorderType)b slotDraggedFrom:(int)fromPos to:(int)toPos
	{
	MiscBorderType const ob = otherBorder(b);
	if ([self autoSortSlots:ob])
		{
		int slot = [self border:b slotAtPosition:toPos];
		if ([self border:b slotSortType:slot] != MISC_SORT_SKIP ||
			[self border:b slotSortFunc:slot] != 0)
			[self sortSlots:ob]; // Don't resort if it doesn't affect ordering.
		}
	id del = [self responsibleDelegate:DF::DEL_SLOT_DRAGGED];
	if (del != 0)
		[del tableScroll:self border:b slotDraggedFrom:fromPos to:toPos];
	}

- (void)border:(MiscBorderType)b slotSortReversed:(int)n
	{
	MiscBorderType const ob = otherBorder(b);
	if ([self autoSortSlots:ob])
		[self sortSlots:ob];			// Assume it affects sorting.
	id del = [self responsibleDelegate:DF::DEL_SLOT_REVERSED];
	if (del != 0)
		{
		int const phys = [self border:b slotAtPosition:n];
		[del tableScroll:self border:b slotSortReversed:phys];
		}
	}

- (void)border:(MiscBorderType)b slotResized:(int)n
	{
	id del = [self responsibleDelegate:DF::DEL_SLOT_RESIZED];
	if (del != 0)
		{
		int const phys = [self border:b slotAtPosition:n];
		[del tableScroll:self border:b slotResized:phys];
		}
	}


//-----------------------------------------------------------------------------
// Target / Action
//-----------------------------------------------------------------------------
- (id)target						{ return target; }
- (id)setTarget:(id)obj				{ target = obj; return self; }
- (id)doubleTarget					{ return doubleTarget; }
- (id)setDoubleTarget:(id)obj		{ doubleTarget = obj; return self; }
- (SEL)action						{ return action; }
- (id)setAction:(SEL)new_sel		{ action = new_sel; return self; }
- (id)setDoubleAction:(SEL)new_sel	{ doubleAction = new_sel; return self; }
- (SEL)doubleAction					{ return doubleAction; }

- (id)sendAction:(SEL)aSel to:(id)obj
		{
		if (aSel == 0)
			aSel = action;
		if (obj == 0)
			obj = target;
		return ([NXApp sendAction:aSel to:obj from:self] ? self : 0);
		}

- (id)sendAction
		{ return [self sendAction:action to:target]; }

- (id)sendDoubleAction
		{ return [self sendAction:doubleAction to:doubleTarget]; }

- (id)sendActionIfEnabled
		{
		if ([self isEnabled])
			return [self sendAction];
		return 0;
		}

- (id)sendDoubleActionIfEnabled
		{
		if ([self isEnabled])
			return [self sendDoubleAction];
		return 0;
		}


//-----------------------------------------------------------------------------
// FONT
//-----------------------------------------------------------------------------
+ (Font*)defaultFont
	{
	return [Font userFontOfSize:12.0 matrix:NX_FLIPPEDMATRIX];
	}

- (id)font { return font; }


static double get_height( Font* font )
	{
	double const LINE_SPACING = 1.20;
	double const size = [font pointSize];
	return size * LINE_SPACING;
	}


- (id)setFont:(id)newFont
	{
	if (newFont != 0 && newFont != (id)font && [newFont isKindOf:[Font class]])
		{
		[window disableDisplay];
		Font* oldFont = font;
		font = newFont;
		NXCoord old_size = [self uniformSizeRows];
		if (old_size != 0)
			{
			// FIXME: Handle this better.  Different cell-types will have
			// different amounts of "fixed-size" border stuff.
			NXCoord const BORDER_THICKNESS = 1;
			old_size -= BORDER_THICKNESS;
			NXCoord const new_size =
				floor(	0.5 + (double(old_size) *
						(get_height( newFont ) / get_height( oldFont ))));
			if (new_size != old_size)
				[self setUniformSizeRows:(new_size + BORDER_THICKNESS)];
			}
		// FIXME: Set font in all existing prototype cells.
		// WARNING: Currently, just asking the border for a prototype
		// in a given slot allocates and initializes an array of prototypes.
		if (![self isLazy])				// Eager beaver sets all cells now.
			{
			int const NRows = num_rows;
			int const NCols = num_cols;
			for (int r = 0;	 r < NRows;	 r++)
				{
				for (int c = 0;	 c < NCols;	 c++)
					{
					id cell = [self cellAt:r:c];
					if (cell != 0)
						{
						if ([cell respondsTo:@selector(setOwnerFont:)])
							[cell setOwnerFont:newFont];
						else if ([cell respondsTo:@selector(setFont:)])
							[cell setFont:newFont];
						}
					}
				}
			}
		id del = [self responsibleDelegate:DF::DEL_FONT_CHANGED];
		if (del != 0)
			[del tableScroll:self fontChangedFrom:oldFont to:newFont];

		[window reenableDisplay];
		[self update];
		}
	return self;
	}


- (id)changeFont:(id)sender
	{
	FontManager* fontMgr = [FontManager new];
	Font* newFont = [fontMgr convertFont:[fontMgr selFont]];
	if (newFont != 0 && newFont != font)
		{
		[self suspendEditing];
		[window disableDisplay];
		Font* oldFont = font;
		[self setFont:newFont];
		id del = [self responsibleDelegate:DF::DEL_CHANGE_FONT];
		if (del != 0)
			[del tableScroll:self changeFont:oldFont to:newFont];
		if (editInfo.editing)
			[editInfo.cell setFont:newFont];
		[window reenableDisplay];
		[self update];
		[self resumeEditing];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// COLOR
//
// The following macros expand into the implementations for these functions.
// Their names are listed so they can be found when searched for.
// setBackgroundColor:, setTextColor:,
// setHighlightBackgroundColor:, setHighlightTextColor:
// setBackgroundGray:, setTextGray:,
// setHighlightBackgroundGray:, setHighlightTextGray:
// backgroundGray:, textGray:, highlightBackgroundGray:, highlightTextGray:
//-----------------------------------------------------------------------------
+ (NXColor)defaultBackgroundColor				{ return NX_COLORLTGRAY; }
+ (NXColor)defaultTextColor						{ return NX_COLORBLACK; }
+ (NXColor)defaultHighlightBackgroundColor		{ return NX_COLORWHITE; }
+ (NXColor)defaultHighlightTextColor			{ return NX_COLORBLACK; }

- (NXColor)backgroundColor				{ return backgroundColor; }
- (NXColor)textColor					{ return textColor; }
- (NXColor)highlightBackgroundColor		{ return highlightBackgroundColor; }
- (NXColor)highlightTextColor			{ return highlightTextColor; }

- (id)setColor:(NXColor)value	{ return [self setBackgroundColor:value]; }
- (NXColor)color				{ return [self backgroundColor]; }

- (id)setColor:(NXColor)x		// New color value.
		var:(NXColor*)v			// Instance variable for the color.
		sel1:(SEL)sel1			// "setOwner...Color:" message for cell.
		sel2:(SEL)sel2			// "set...Color:" message for cell.
		notifySel:(MiscDelegateFlags::Selector)notifySel // "..ColorChangedTo:"
	{
	if (!NXEqualColor( x, *v ))
		{
		BOOL const was_auto = [self isAutodisplay];
		[self setAutodisplay:NO];

		*v = x;
		if (v == &backgroundColor) [self forwardBGColor];

		if (![self isLazy])
			{
			int const NRows = num_rows;
			int const NCols = num_cols;
			for (int r = 0; r < NRows;	r++)
				for (int c = 0;	 c < NCols;	 c++)
					{
					id cell = [self cellAt:r:c];
					if (cell != 0)
						{
						if ([cell respondsTo:sel1])
							(*[cell methodFor:sel1])( cell, sel1, x );
						else if ([cell respondsTo:sel2])
							(*[cell methodFor:sel2])( cell, sel2, x );
						}
					}
			}

		id del = [self responsibleDelegate:notifySel];
		if (del != 0)
			{
			SEL objcSel = MiscDelegateFlags::selToObjc( notifySel );
			(*[del methodFor:objcSel])( del, objcSel, self, x );
			}

		[self setNeedsDisplay:YES];
		[self setAutodisplay:was_auto];			// Will display if needed.
		}

	return self;
	}


#define MISC_SET_COLOR_FUNC(LNAME,CNAME,UNAME)\
- (id)set##CNAME##Color:(NXColor) value\
	{ return [self setColor:value\
				var:& LNAME##Color\
				sel1:@selector(setOwner##CNAME##Color:)\
				sel2:@selector(set##CNAME##Color:)\
				notifySel:MiscDelegateFlags::DEL_##UNAME##_COLOR_CHANGED]; }

MISC_SET_COLOR_FUNC( background, Background, BACK )
MISC_SET_COLOR_FUNC( text, Text, TEXT )
MISC_SET_COLOR_FUNC( highlightBackground, HighlightBackground, BACKH )
MISC_SET_COLOR_FUNC( highlightText, HighlightText, TEXTH )
#undef MISC_SET_COLOR_FUNC


static inline float color_to_gray( NXColor x )
	{
	float gray;
	NXConvertColorToGray( x, &gray );
	return gray;
	}


#define MISC_GET_GRAY_FUNC( LNAME )\
- (float)LNAME##Gray { return color_to_gray( [self LNAME##Color] ); }

MISC_GET_GRAY_FUNC( background )
MISC_GET_GRAY_FUNC( text )
MISC_GET_GRAY_FUNC( highlightBackground )
MISC_GET_GRAY_FUNC( highlightText )
#undef MISC_GET_GRAY_FUNC

#define MISC_SET_GRAY_FUNC( CNAME )\
- (id)set##CNAME##Gray:(float)x\
	{ return [self set##CNAME##Color: NXConvertGrayToColor(x)]; }

MISC_SET_GRAY_FUNC( Background )
MISC_SET_GRAY_FUNC( Text )
MISC_SET_GRAY_FUNC( HighlightBackground )
MISC_SET_GRAY_FUNC( HighlightText )
#undef MISC_SET_GRAY_FUNC


//=============================================================================
// SAVE / RESTORE
//=============================================================================

//-----------------------------------------------------------------------------
// SLOT ORDER
//-----------------------------------------------------------------------------
- (id)border:(MiscBorderType)b slotOrder:(MiscIntList*)list
	{
	if (list != 0)
		{
		MiscTableBorder const* const bp = info[b]->border;
		MiscCoord_P const* const vmap = bp->getP2VMap();
		int const lim = bp->count();
		[list empty];
		for (int i = 0;	 i < lim;  i++)
			{
			int v = (vmap ? vmap[i] : i);
			if (bp->getSortDirection(v) == MISC_SORT_DESCENDING)
				v = ~v;
			[list addInt:v];
			}
		return self;
		}
	return 0;
	}
- (id)colOrder:(MiscIntList*)list
	{ return [self border:MISC_COL_BORDER slotOrder:list]; }
- (id)rowOrder:(MiscIntList*)list
	{ return [self border:MISC_ROW_BORDER slotOrder:list]; }

- (id)border:(MiscBorderType)b setSlotOrder:(MiscIntList*)list
	{
	id ret = 0;
	MiscTableBorder* const bp = info[b]->border;
	if (list != 0)
		{
		int const lim = bp->count();
		if ([list count] == lim)
			{
			unsigned int const nbytes = lim * sizeof(int) + lim * sizeof(bool);
			int* const map = (int*) malloc( nbytes );
			MiscAssert( map != 0 );
			bool* const desc = (bool*) (map + lim);

			for (int i = 0; i < lim; i++)
				{
				int const j = [list intAt:i];
				map[i] = (desc[i] = (j < 0)) ? ~j : j;
				}

			if (bp->setP2VMap( map ))
				{
				for (int i = 0; i < lim; i++)
					{
					MiscSortDirection const dir =
						(desc[i] ? MISC_SORT_DESCENDING : MISC_SORT_ASCENDING);
					if (dir != bp->getSortDirection( map[i] ))
						bp->setSortDirection( map[i], dir );
					}

				MiscBorderType const ob = otherBorder(b);
				if ([self autoSortSlots:ob])
					[self sortSlots:ob];
				ret = self;
				}

			free( map );
			}
		}
	else
		{
		bp->setP2VMap( 0 );
		bp->clearSortDirection();
		ret = self;
		}

	return ret;
	}

- (id)setColOrder:(MiscIntList*)list
	{ return [self border:MISC_COL_BORDER setSlotOrder:list]; }
- (id)setRowOrder:(MiscIntList*)list
	{ return [self border:MISC_ROW_BORDER setSlotOrder:list]; }

- (char*)border:(MiscBorderType)b slotOrderAsString:(char*)buff
		size:(int)buff_size canExpand:(BOOL)canExpand
	{
	char* retval;
	MiscIntList* list = [[MiscIntList alloc] init];
	[self border:b slotOrder:list];
	retval = [list writeToString:buff size:buff_size canExpand:canExpand];
	[list free];
	return retval;
	}
- (char*)colOrderAsString:(char*)buff
		size:(int)buff_size canExpand:(BOOL)canExpand
	{ return [self border:MISC_COL_BORDER slotOrderAsString:buff
						size:buff_size canExpand:canExpand]; }
- (char*)rowOrderAsString:(char*)buff
		size:(int)buff_size canExpand:(BOOL)canExpand
	{ return [self border:MISC_ROW_BORDER slotOrderAsString:buff
						size:buff_size canExpand:canExpand]; }

- (id)border:(MiscBorderType)b setSlotOrderFromString:(char const*)s
	{
	id retval;
	MiscIntList* list = [[MiscIntList alloc] init];
	[list readFromString:s];
	retval = [self border:b setSlotOrder:list];
	[list free];
	return retval;
	}
- (id)setColOrderFromString:(char const*)s
	{ return [self border:MISC_COL_BORDER setSlotOrderFromString:s]; }
- (id)setRowOrderFromString:(char const*)s
	{ return [self border:MISC_ROW_BORDER setSlotOrderFromString:s]; }


//-----------------------------------------------------------------------------
// SLOT SIZES
//-----------------------------------------------------------------------------
- (id)border:(MiscBorderType)b slotSizes:(MiscIntList*)list
	{
	if (list != 0)
		{
		MiscTableBorder const* const bp = info[b]->border;
		int const lim = bp->count();
		[list empty];
		for (int i = 0;	 i < lim;  i++)
			[list addInt: (int) bp->getSize_P( i ) ];
		return self;
		}
	return 0;
	}
- (id)colSizes:(MiscIntList*)list
	{ return [self border:MISC_COL_BORDER slotSizes:list]; }
- (id)rowSizes:(MiscIntList*)list
	{ return [self border:MISC_ROW_BORDER slotSizes:list]; }

- (id)border:(MiscBorderType)b setSlotSizes:(MiscIntList*)list
	{
	if (list != 0)
		{
		MiscTableBorder* const bp = info[b]->border;
		int const lim = bp->count();
		if ([list count] == lim)
			{
			for (int i = 0;	 i < lim;  i++)
				{
				NXCoord const min_size = [self border:b slotMinSize:i];
				NXCoord const max_size = [self border:b slotMaxSize:i];
				int s = [list intAt:i];
				if (s < min_size)
					s = (int)min_size;
				else if (s > max_size)
					s = (int)max_size;
				bp->setSize_P( i, (MiscPixels)s );
				}
			[self constrainSize];
			return self;
			}
		}
	return 0;
	}
- (id)setColSizes:(MiscIntList*)list
	{ return [self border:MISC_COL_BORDER setSlotSizes:list]; }
- (id)setRowSizes:(MiscIntList*)list
	{ return [self border:MISC_ROW_BORDER setSlotSizes:list]; }

- (char*)border:(MiscBorderType)b slotSizesAsString:(char*)buff
		size:(int)buff_size canExpand:(BOOL)canExpand
	{
	char* retval;
	MiscIntList* list = [[MiscIntList alloc] init];
	[self border:b slotSizes:list];
	retval = [list writeToString:buff size:buff_size canExpand:canExpand];
	[list free];
	return retval;
	}
- (char*)colSizesAsString:(char*)buff
		size:(int)buff_size canExpand:(BOOL)canExpand
	{ return [self border:MISC_COL_BORDER slotSizesAsString:buff
						size:buff_size canExpand:canExpand]; }
- (char*)rowSizesAsString:(char*)buff
		size:(int)buff_size canExpand:(BOOL)canExpand
	{ return [self border:MISC_ROW_BORDER slotSizesAsString:buff
						size:buff_size canExpand:canExpand]; }

- (id)border:(MiscBorderType)b setSlotSizesFromString:(char const*)s
	{
	id retval;
	MiscIntList* list = [[MiscIntList alloc] init];
	[list readFromString:s];
	retval = [self border:b setSlotSizes:list];
	[list free];
	return retval;
	}
- (id)setColSizesFromString:(char const*)s
	{ return [self border:MISC_COL_BORDER setSlotSizesFromString:s]; }
- (id)setRowSizesFromString:(char const*)s
	{ return [self border:MISC_ROW_BORDER setSlotSizesFromString:s]; }



//-----------------------------------------------------------------------------
// Border Views
//-----------------------------------------------------------------------------
- (MiscCornerView*)cornerView			{ return cornerView; }
- (char const*)cornerTitle				{ return [cornerView title]; }
- (void)setCornerTitle:(char const*)s	{ [cornerView setTitle:s]; }

- (BOOL)setBorder:(MiscBorderType)type on:(BOOL)on
	{
	MiscBorderInfo& b = *(info[type]);
	if (b.isOn != on)
		{
		[window disableDisplay];
		BOOL const other_border_is_on = info[otherBorder(type)]->isOn;
		b.isOn = on;
		if (on)
			{
			if (other_border_is_on)
				[self addSubview: cornerView];
			[self addSubview: b.clip];
			NXRect r,c;
			[b.clip getBounds:&r];
			[contentView getBounds:&c];
			if (type == MISC_COL_BORDER)
				{
				if (r.origin.x != c.origin.x)
					{ r.origin.x = c.origin.x; [b.clip rawScroll:&r.origin]; }
				}
			else
				{
				if (r.origin.y != c.origin.y)
					{ r.origin.y = c.origin.y; [b.clip rawScroll:&r.origin]; }
				}
			}
		else
			{
			[b.clip removeFromSuperview];
			if (other_border_is_on)
				[cornerView removeFromSuperview];
			}
		[self tile];
		[self constrainSize];
		[window reenableDisplay];
		[self update];
		return YES;
		}
	return NO;
	}


//-----------------------------------------------------------------------------
// SLOT methods
//-----------------------------------------------------------------------------
- (MiscTableBorder*)border:(MiscBorderType)b
		{ return info[b]->border; }
- (BOOL)border:(MiscBorderType)b setTitlesOn:(BOOL)on_off
		{ return [self setBorder:b on:on_off]; }
- (BOOL)borderTitlesOn:(MiscBorderType)b
		{ return info[b]->isOn; }
- (MiscTableTitleMode)borderTitleMode:(MiscBorderType)b
		{ return info[b]->border->getTitleMode(); }
- (void)border:(MiscBorderType)b setTitleMode:(MiscTableTitleMode)x
		{
		MiscBorderInfo* const ip = info[b];
		if (ip->border->setTitleMode(x) && ip->isOn && ip->border->count() > 0)
			[ip->view update];
		}
- (NXCoord)borderTitlesSize:(MiscBorderType)b
		{ return info[b]->isOn ? [info[b]->view frameHeight] : 0; }
- (void)border:(MiscBorderType)b setTitlesSize:(NXCoord)x
		{
		[info[b]->view setFrameHeight:(MiscPixels)floor(x)];
		[self tile];
		}

- (id)border:(MiscBorderType)b moveSlotFrom:(int)fromPos to:(int)toPos
		{
		info[b]->border->moveFromTo(fromPos,toPos);
		[self selectionChanged];
		return self;
		}
- (int)border:(MiscBorderType)b slotPosition:(int)n
		{ return info[b]->border->physicalToVisual(n); }
- (int)border:(MiscBorderType)b slotAtPosition:(int)n
		{ return info[b]->border->visualToPhysical(n); }

- (void)border:(MiscBorderType)b physicalToVisual:(MiscIntList*)list
		{
		for (unsigned int i = [list count];	 i-- > 0; )
			{
			MiscCoord_P const p = [list intAt:i];
			[list replaceIntAt:i with:[self border:b slotPosition:p]];
			}
		}

- (void)border:(MiscBorderType)b visualToPhysical:(MiscIntList*)list
		{
		for (unsigned int i = [list count];	 i-- > 0; )
			{
			MiscCoord_V const v = [list intAt:i];
			[list replaceIntAt:i with:[self border:b slotAtPosition:v]];
			}
		}


- (BOOL)sizeableSlots:(MiscBorderType)b
		{ return info[b]->border->isSizeable(); }
- (BOOL)draggableSlots:(MiscBorderType)b
		{ return info[b]->border->isDraggable(); }
- (BOOL)modifierDragSlots:(MiscBorderType)b
		{ return info[b]->border->isModifierDrag(); }
- (NXCoord)uniformSizeSlots:(MiscBorderType)b
		{ return (NXCoord) info[b]->border->getUniformSize(); }
- (NXCoord)minUniformSizeSlots:(MiscBorderType)b
		{ return (NXCoord) info[b]->border->getMinUniformSize(); }
- (NXCoord)maxUniformSizeSlots:(MiscBorderType)b
		{ return (NXCoord) info[b]->border->getMaxUniformSize(); }

- (NXCoord)border:(MiscBorderType)b slotAdjustedSize:(int)n
		{ return (NXCoord) info[b]->border->effectiveSize_P(n); }
- (NXCoord)border:(MiscBorderType)b slotSize:(int)n
		{ return (NXCoord) info[b]->border->getSize_P(n); }
- (NXCoord)border:(MiscBorderType)b slotMinSize:(int)n
		{ return (NXCoord) info[b]->border->getMinSize_P(n); }
- (NXCoord)border:(MiscBorderType)b slotMaxSize:(int)n
		{ return (NXCoord) info[b]->border->getMaxSize_P(n); }
- (BOOL)border:(MiscBorderType)b slotIsSizeable:(int)n
		{ return info[b]->border->isSizeable_P(n); }
- (BOOL)border:(MiscBorderType)b slotIsAutosize:(int)n
		{ return info[b]->border->isSpringy_P(n); }
- (char const*)border:(MiscBorderType)b slotTitle:(int)n
		{ return info[b]->border->getTitle_P(n); }
- (int)border:(MiscBorderType)b slotTag:(int)n
		{ return info[b]->border->getTag_P(n); }
- (MiscTableCellStyle)border:(MiscBorderType)b slotCellType:(int)n
		{ return info[b]->border->getStyle_P(n); }
- (id)border:(MiscBorderType)b slotCellPrototype:(int)n
		{ return info[b]->border->getPrototype_P(n); }

- (void)border:(MiscBorderType)b setSizeableSlots:(BOOL)flag
		{ info[b]->border->setSizeable( flag ); }
- (void)border:(MiscBorderType)b setDraggableSlots:(BOOL)flag
		{ info[b]->border->setDraggable( flag ); }
- (void)border:(MiscBorderType)b setModifierDragSlots:(BOOL)flag
		{ info[b]->border->setModifierDrag( flag ); }
- (void)border:(MiscBorderType)b setUniformSizeSlots:(NXCoord)uniform_size
		{
		MiscBorderInfo* const ip = info[b];
		if (ip->border->setUniformSize( (MiscPixels)floor(uniform_size) ))
			{
			[self constrainSize];
			if (b == MISC_ROW_BORDER)
				{
				NXCoord const scr_size = uniform_size != 0 ?
						uniform_size : ip->border->getDefaultSize();
				[self setLineScroll:scr_size];
				[self setPageScroll:scr_size];
				}
			[self update];
			}
		}
- (void)border:(MiscBorderType)b setMinUniformSizeSlots:(NXCoord)size
		{ info[b]->border->setMinUniformSize( (MiscPixels)floor(size) ); }
- (void)border:(MiscBorderType)b setMaxUniformSizeSlots:(NXCoord)size
		{ info[b]->border->setMaxUniformSize( (MiscPixels)floor(size) ); }

- (void)border:(MiscBorderType)b setSlot:(int)n size:(NXCoord)size
		{
		info[b]->border->setSize_P( n, (MiscPixels)floor(size) );
		[self constrainSize];
		// FIXME: [self update];
		}
- (void)border:(MiscBorderType)b setSlot:(int)n minSize:(NXCoord)size
		{
		info[b]->border->setMinSize_P( n, (MiscPixels)floor(size) );
		[self constrainSize];
		}
- (void)border:(MiscBorderType)b setSlot:(int)n maxSize:(NXCoord)size
		{
		info[b]->border->setMaxSize_P( n, (MiscPixels)floor(size) );
		[self constrainSize];
		}
- (void)border:(MiscBorderType)b setSlot:(int)n sizeable:(BOOL)flag
		{ info[b]->border->setSizeable_P( n, flag ); }
- (void)border:(MiscBorderType)b setSlot:(int)n autosize:(BOOL)flag
		{
		info[b]->border->setSpringy_P( n, flag );
		[self constrainSize];
		}

- (void)border:(MiscBorderType)b setSlot:(int)n title:(char const*)title
		{
		MiscBorderInfo* const ip = info[b];
		if (ip->border->setTitle_P( n, title ) && ip->isOn)
			[self border:b drawSlotTitle:n];
		}

- (void)border:(MiscBorderType)b setSlot:(int)n tag:(int)x
		{ info[b]->border->setTag_P( n, x ); }
- (void)border:(MiscBorderType)b setSlot:(int)n
				cellType:(MiscTableCellStyle)type
		{ info[b]->border->setStyle_P(n,type); }
- (void)border:(MiscBorderType)b setSlot:(int)n cellPrototype:(id)p
		{ info[b]->border->setPrototype_P(n,p); }



//-----------------------------------------------------------------------------
// COL methods
//-----------------------------------------------------------------------------
- (MiscTableBorder*)colBorder
		{ return colInfo.border; }
- (BOOL)colTitlesOn
		{ return [self borderTitlesOn:MISC_COL_BORDER]; }
- (BOOL)setColTitlesOn:(BOOL)x
		{ return [self setBorder:MISC_COL_BORDER on:x]; }
- (MiscTableTitleMode)colTitleMode
		{ return [self borderTitleMode:MISC_COL_BORDER]; }
- (void)setColTitleMode:(MiscTableTitleMode)x
		{ [self border:MISC_COL_BORDER setTitleMode:x]; }
- (NXCoord)colTitlesHeight
		{ return [self borderTitlesSize:MISC_COL_BORDER]; }
- (void)setColTitlesHeight:(NXCoord)x
		{ [self border:MISC_COL_BORDER setTitlesSize:x]; }
- (MiscBorderView*)colTitles
		{ return ([self colTitlesOn] ? colInfo.view : 0); }

- (id)moveColFrom:(int)fromPos to:(int)toPos
		{ return [self border:MISC_COL_BORDER moveSlotFrom:fromPos to:toPos]; }
- (int)colPosition:(int)n
		{ return [self border:MISC_COL_BORDER slotPosition:n]; }
- (int)colAtPosition:(int)n
		{ return [self border:MISC_COL_BORDER slotAtPosition:n]; }

- (BOOL)sizeableCols
		{ return [self sizeableSlots:MISC_COL_BORDER]; }
- (BOOL)draggableCols
		{ return [self draggableSlots:MISC_COL_BORDER]; }
- (BOOL)modifierDragCols
		{ return [self modifierDragSlots:MISC_COL_BORDER]; }
- (NXCoord)uniformSizeCols
		{ return [self uniformSizeSlots:MISC_COL_BORDER]; }
- (NXCoord)minUniformSizeCols
		{ return [self minUniformSizeSlots:MISC_COL_BORDER]; }
- (NXCoord)maxUniformSizeCols
		{ return [self maxUniformSizeSlots:MISC_COL_BORDER]; }

- (NXCoord)colAdjustedSize:(int)n
		{ return [self border:MISC_COL_BORDER slotAdjustedSize:n]; }
- (NXCoord)colSize:(int)n
		{ return [self border:MISC_COL_BORDER slotSize:n]; }
- (NXCoord)colMinSize:(int)n
		{ return [self border:MISC_COL_BORDER slotMinSize:n]; }
- (NXCoord)colMaxSize:(int)n
		{ return [self border:MISC_COL_BORDER slotMaxSize:n]; }
- (BOOL)colIsSizeable:(int)n
		{ return [self border:MISC_COL_BORDER slotIsSizeable:n]; }
- (BOOL)colIsAutosize:(int)n
		{ return [self border:MISC_COL_BORDER slotIsAutosize:n]; }
- (char const*)colTitle:(int)n
		{ return [self border:MISC_COL_BORDER slotTitle:n]; }
- (int)colTag:(int)n
		{ return [self border:MISC_COL_BORDER slotTag:n]; }
- (MiscTableCellStyle)colCellType:(int)n
		{ return [self border:MISC_COL_BORDER slotCellType:n]; }
- (id)colCellPrototype:(int)n
		{ return [self border:MISC_COL_BORDER slotCellPrototype:n]; }

- (void)setSizeableCols:(BOOL)flag
		{ [self border:MISC_COL_BORDER setSizeableSlots:flag]; }
- (void)setDraggableCols:(BOOL)flag
		{ [self border:MISC_COL_BORDER setDraggableSlots:flag]; }
- (void)setModifierDragCols:(BOOL)flag
		{ [self border:MISC_COL_BORDER setModifierDragSlots:flag]; }
- (void)setUniformSizeCols:(NXCoord)size
		{ [self border:MISC_COL_BORDER setUniformSizeSlots:size]; }
- (void)setMinUniformSizeCols:(NXCoord)size
		{ [self border:MISC_COL_BORDER setMinUniformSizeSlots:size]; }
- (void)setMaxUniformSizeCols:(NXCoord)size
		{ [self border:MISC_COL_BORDER setMaxUniformSizeSlots:size]; }

- (void)setCol:(int)n size:(NXCoord)size
		{ [self border:MISC_COL_BORDER setSlot:n size:size]; }
- (void)setCol:(int)n minSize:(NXCoord)size
		{ [self border:MISC_COL_BORDER setSlot:n minSize:size]; }
- (void)setCol:(int)n maxSize:(NXCoord)size
		{ [self border:MISC_COL_BORDER setSlot:n maxSize:size]; }
- (void)setCol:(int)n sizeable:(BOOL)flag
		{ [self border:MISC_COL_BORDER setSlot:n sizeable:flag]; }
- (void)setCol:(int)n autosize:(BOOL)flag
		{ [self border:MISC_COL_BORDER setSlot:n autosize:flag]; }
- (void)setCol:(int)n title:(char const*)title
		{ [self border:MISC_COL_BORDER setSlot:n title:title]; }
- (void)setCol:(int)n tag:(int)x
		{ [self border:MISC_COL_BORDER setSlot:n tag:x]; }
- (void)setCol:(int)n cellType:(MiscTableCellStyle)x
		{ [self border:MISC_COL_BORDER setSlot:n cellType:x]; }
- (void)setCol:(int)n cellPrototype:p
		{ [self border:MISC_COL_BORDER setSlot:n cellPrototype:p]; }


//-----------------------------------------------------------------------------
// ROW methods
//-----------------------------------------------------------------------------
- (MiscTableBorder*)rowBorder
		{ return rowInfo.border; }
- (BOOL)rowTitlesOn
		{ return [self borderTitlesOn:MISC_ROW_BORDER]; }
- (BOOL)setRowTitlesOn:(BOOL)x
		{ return [self setBorder:MISC_ROW_BORDER on:x]; }
- (MiscTableTitleMode)rowTitleMode
		{ return [self borderTitleMode:MISC_ROW_BORDER]; }
- (void)setRowTitleMode:(MiscTableTitleMode)x
		{ [self border:MISC_ROW_BORDER setTitleMode:x]; }
- (NXCoord)rowTitlesWidth
		{ return [self borderTitlesSize:MISC_ROW_BORDER]; }
- (void)setRowTitlesWidth:(NXCoord)x
		{ [self border:MISC_ROW_BORDER setTitlesSize:x]; }
- (MiscBorderView*)rowTitles
		{ return ([self rowTitlesOn] ? rowInfo.view : 0); }

- (id)moveRowFrom:(int)fromPos to:(int)toPos
		{ return [self border:MISC_ROW_BORDER moveSlotFrom:fromPos to:toPos]; }
- (int)rowPosition:(int)n
		{ return [self border:MISC_ROW_BORDER slotPosition:n]; }
- (int)rowAtPosition:(int)n
		{ return [self border:MISC_ROW_BORDER slotAtPosition:n]; }

- (BOOL)sizeableRows
		{ return [self sizeableSlots:MISC_ROW_BORDER]; }
- (BOOL)draggableRows
		{ return [self draggableSlots:MISC_ROW_BORDER]; }
- (BOOL)modifierDragRows
		{ return [self modifierDragSlots:MISC_ROW_BORDER]; }
- (NXCoord)uniformSizeRows
		{ return [self uniformSizeSlots:MISC_ROW_BORDER]; }
- (NXCoord)minUniformSizeRows
		{ return [self minUniformSizeSlots:MISC_ROW_BORDER]; }
- (NXCoord)maxUniformSizeRows
		{ return [self maxUniformSizeSlots:MISC_ROW_BORDER]; }

- (NXCoord)rowAdjustedSize:(int)n
		{ return [self border:MISC_ROW_BORDER slotAdjustedSize:n]; }
- (NXCoord)rowSize:(int)n
		{ return [self border:MISC_ROW_BORDER slotSize:n]; }
- (NXCoord)rowMinSize:(int)n
		{ return [self border:MISC_ROW_BORDER slotMinSize:n]; }
- (NXCoord)rowMaxSize:(int)n
		{ return [self border:MISC_ROW_BORDER slotMaxSize:n]; }
- (BOOL)rowIsSizeable:(int)n
		{ return [self border:MISC_ROW_BORDER slotIsSizeable:n]; }
- (BOOL)rowIsAutosize:(int)n
		{ return [self border:MISC_ROW_BORDER slotIsAutosize:n]; }
- (char const*)rowTitle:(int)n
		{ return [self border:MISC_ROW_BORDER slotTitle:n]; }
- (int)rowTag:(int)n
		{ return [self border:MISC_ROW_BORDER slotTag:n]; }
- (MiscTableCellStyle)rowCellType:(int)n
		{ return [self border:MISC_ROW_BORDER slotCellType:n]; }
- (id)rowCellPrototype:(int)n
		{ return [self border:MISC_ROW_BORDER slotCellPrototype:n]; }

- (void)setSizeableRows:(BOOL)flag
		{ [self border:MISC_ROW_BORDER setSizeableSlots:flag]; }
- (void)setDraggableRows:(BOOL)flag
		{ [self border:MISC_ROW_BORDER setDraggableSlots:flag]; }
- (void)setModifierDragRows:(BOOL)flag
		{ [self border:MISC_ROW_BORDER setModifierDragSlots:flag]; }
- (void)setUniformSizeRows:(NXCoord)size
		{ [self border:MISC_ROW_BORDER setUniformSizeSlots:size]; }
- (void)setMinUniformSizeRows:(NXCoord)size
		{ [self border:MISC_ROW_BORDER setMinUniformSizeSlots:size]; }
- (void)setMaxUniformSizeRows:(NXCoord)size
		{ [self border:MISC_ROW_BORDER setMaxUniformSizeSlots:size]; }

- (void)setRow:(int)n size:(NXCoord)size
		{ [self border:MISC_ROW_BORDER setSlot:n size:size]; }
- (void)setRow:(int)n minSize:(NXCoord)size
		{ [self border:MISC_ROW_BORDER setSlot:n minSize:size]; }
- (void)setRow:(int)n maxSize:(NXCoord)size
		{ [self border:MISC_ROW_BORDER setSlot:n maxSize:size]; }
- (void)setRow:(int)n sizeable:(BOOL)flag
		{ [self border:MISC_ROW_BORDER setSlot:n sizeable:flag]; }
- (void)setRow:(int)n autosize:(BOOL)flag
		{ [self border:MISC_ROW_BORDER setSlot:n autosize:flag]; }
- (void)setRow:(int)n title:(char const*)title
		{ [self border:MISC_ROW_BORDER setSlot:n title:title]; }
- (void)setRow:(int)n tag:(int)x
		{ [self border:MISC_ROW_BORDER setSlot:n tag:x]; }
- (void)setRow:(int)n cellType:(MiscTableCellStyle)x
		{ [self border:MISC_ROW_BORDER setSlot:n cellType:x]; }
- (void)setRow:(int)n cellPrototype:p
		{ [self border:MISC_ROW_BORDER setSlot:n cellPrototype:p]; }


//-----------------------------------------------------------------------------
// DRAWING
//		FIXME: This should all be routed through the -display::: mechanism
//		so that the lockFocus can be done automatically, and so that
//		subviews will have an opportunity to draw themselves.
//-----------------------------------------------------------------------------

- (id)drawCellAt:(int)row :(int)col
	{
	if ([tableView canDraw])
		{
		[tableView lockFocus];
		[tableView drawCellAt:row:col];
		[tableView unlockFocus];
		[window flushWindow];
		}
	return self;
	}

- (id)drawRow:(int)row
	{
	if ([tableView canDraw])
		{
		[tableView lockFocus];
		[tableView drawRow:row];
		[tableView unlockFocus];
		[window flushWindow];
		}
	return self;
	}

- (id)drawCol:(int)col
	{
	if ([tableView canDraw])
		{
		[tableView lockFocus];
		[tableView drawCol:col];
		[tableView unlockFocus];
		[window flushWindow];
		}
	return self;
	}

- (id)border:(MiscBorderType)b drawSlot:(int)n
	{ return (b == MISC_COL_BORDER) ? [self drawCol:n] : [self drawRow:n]; }

- (id)border:(MiscBorderType)b drawSlotTitle:(int)n		// physical position.
	{
	MiscBorderInfo* const ip = info[b];
	if (ip->isOn)										// visual position.
		{
		MiscBorderView* const v= ip->view;
		if ([v canDraw])
			{
			[v lockFocus];
			[v drawSlot: [self border:b slotPosition:n]];
			[v unlockFocus];
			[window flushWindow];
			}
		}
	return self;
	}
- (id)drawRowTitle:(int)n
		{ return [self border:MISC_ROW_BORDER drawSlotTitle:n]; }
- (id)drawColTitle:(int)n
		{ return [self border:MISC_COL_BORDER drawSlotTitle:n]; }

- (id)setAutodisplay:(BOOL)x
	{
	[rowInfo.view setAutodisplay:x];
	[colInfo.view setAutodisplay:x];
	[tableView setAutodisplay:x];
	return [super setAutodisplay:x];
	}

- (BOOL)drawClippedText					{ return drawClippedText; }
- (void)setDrawClippedText:(BOOL)x
	{
	if (drawClippedText != x)
		{
		drawClippedText = x;
		[self update];
		}
	}

//-----------------------------------------------------------------------------
// VISIBLE / SCROLLING
//-----------------------------------------------------------------------------

- (void)scrollCellToVisible:(int)row :(int)col
		{ [tableView scrollCellToVisible: row:col]; }
- (void)scrollRowToVisible:(int)row { [tableView scrollRowToVisible:row]; }
- (void)scrollColToVisible:(int)col { [tableView scrollColToVisible:col]; }
- (id)scrollSelToVisible
	{
	if ([self hasRowSelection])
		[self scrollRowToVisible: [self selectedRow]];
	else if ([self hasColSelection])
		[self scrollColToVisible: [self selectedCol]];
	return self;
	}


- (int)numVisibleSlots:(MiscBorderType)b
		{ return [tableView numVisibleSlots:b]; }
- (int)firstVisibleSlot:(MiscBorderType)b
		{ return [tableView firstVisibleSlot:b]; }
- (int)lastVisibleSlot:(MiscBorderType)b
		{ return [tableView lastVisibleSlot:b]; }
- (BOOL)border:(MiscBorderType)b slotIsVisible:(int)n
		{ return [tableView border:b slotIsVisible:n]; }
- (void)border:(MiscBorderType)b setFirstVisibleSlot:(int)n
		{ [tableView border:b setFirstVisibleSlot:n]; }
- (void)border:(MiscBorderType)b setLastVisibleSlot:(int)n
		{ [tableView border:b setLastVisibleSlot:n]; }

- (int)numVisibleCols  { return [self numVisibleSlots: MISC_COL_BORDER]; }
- (int)firstVisibleCol { return [self firstVisibleSlot:MISC_COL_BORDER]; }
- (int)lastVisibleCol  { return [self lastVisibleSlot: MISC_COL_BORDER]; }
- (BOOL)colIsVisible:(int)n
		{ return [self border:MISC_COL_BORDER slotIsVisible:n]; }
- (void)setFirstVisibleCol:(int)n
		{ [self border:MISC_COL_BORDER setFirstVisibleSlot:n]; }
- (void)setLastVisibleCol:(int)n
		{ [self border:MISC_COL_BORDER setLastVisibleSlot:n]; }

- (int)numVisibleRows  { return [self numVisibleSlots: MISC_ROW_BORDER]; }
- (int)firstVisibleRow { return [self firstVisibleSlot:MISC_ROW_BORDER]; }
- (int) lastVisibleRow { return [self lastVisibleSlot: MISC_ROW_BORDER]; }
- (BOOL)rowIsVisible:(int)n
		{ return [self border:MISC_ROW_BORDER slotIsVisible:n]; }
- (void)setFirstVisibleRow:(int)n
		{ [self border:MISC_ROW_BORDER setFirstVisibleSlot:n]; }
- (void)setLastVisibleRow:(int)n
		{ [self border:MISC_ROW_BORDER setLastVisibleSlot:n]; }
@end
