//=============================================================================
//
//	Copyright (C) 1995,1996,1997,1998 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableScrollIO.M
//
//		Input and Output (IO) methods for MiscTableScroll.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableScrollIO.M,v 1.2 98/03/24 04:47:32 sunshine Exp $
// $Log:		MiscTableScrollIO.M,v $
//	Revision 1.2  98/03/24	04:47:32  sunshine
//	v138: Removed many unnecessary header imports.
//	
//	Revision 1.1  97/12/19	01:06:28  zarnuk
//	v133: Broke off MiscTableScrollIO.M.  New storage layout.
//-----------------------------------------------------------------------------
#import <misckit/MiscTableScroll.h>
#import "MiscCornerView.h"
#import "MiscDelegateFlags.h"
#import "MiscTableBorder.h"
#import "MiscTableScrollPrivate.h"
#import "MiscTableView.h"
#import <new.h>
extern "Objective-C" {
#import <appkit/ClipView.h>
}

@implementation MiscTableScroll(IO)

//-----------------------------------------------------------------------------
// - awakeFromNib
//-----------------------------------------------------------------------------
- (id)awakeFromNib
	{
	delegateFlags->setDelegate( [self delegate] );
	dataDelegateFlags->setDelegate( [self dataDelegate] );
	return self;
	}


//-----------------------------------------------------------------------------
// - read:selector:
//-----------------------------------------------------------------------------
- (SEL)read:(int)ver selector:(NXTypedStream*)stream
	{
	SEL cmd = 0;
	char* str;
	NXReadType( stream, @encode(char*), &str );
	if (str != 0)
		{
		if (*str != '\0')
			cmd = sel_getUid( str );
		free( str );
		}
	return cmd;
	}


//-----------------------------------------------------------------------------
// - read:cornerTitle:globalInfo:
//		NOTE: Cannot archive sort_entry_func and sort_slot_func because they 
//				are function addresses.	 
//-----------------------------------------------------------------------------
- (void)read:(int)ver cornerTitle:(char**)p
		globalInfo:(NXTypedStream*)stream
	{
	NXReadType( stream, @encode(int), &tag );
	NXReadType( stream, @encode(BOOL), &enabled );
	NXReadType( stream, @encode(BOOL), &lazy );
	if (ver < MISC_TS_VERSION_2)
		drawClippedText = NO;
	else
		NXReadType( stream, @encode(BOOL), &drawClippedText );
	NXReadType( stream, @encode(MiscSelectionMode), &mode );
	font = NXReadObject( stream );
	textColor = NXReadColor( stream );
	backgroundColor = NXReadColor( stream );
	highlightTextColor = NXReadColor( stream );
	highlightBackgroundColor = NXReadColor( stream );
	nextText	= NXReadObject( stream );
	previousText= NXReadObject( stream );
	delegate	= NXReadObject( stream );
	dataDelegate= NXReadObject( stream );
	target		= NXReadObject( stream );
	doubleTarget= NXReadObject( stream );
	action = [self read:ver selector:stream];
	doubleAction = [self read:ver selector:stream];
	if (ver < MISC_TS_VERSION_2)
		*p = 0;
	else
		NXReadType( stream, @encode(char*), p );
	}


//-----------------------------------------------------------------------------
// - read:border:stream:
//-----------------------------------------------------------------------------
- (void)read:(int)ver border:(MiscBorderInfo*)p stream:(NXTypedStream*)stream
	{
	NXZone* const z = [self zone];
	int n;
	if (p->sort_vector != 0)
		NXZoneFree( z, p->sort_vector );
	NXReadType( stream, @encode(int), &(n) );
	if (n > 0)
		{
		p->sort_vector_len = n;
		p->sort_vector = (int*) NXZoneMalloc( z, n * sizeof(int) );
		for (int i = 0;	 i < n;	 i++)
			NXReadType( stream, @encode(int), &(p->sort_vector[i]) );
		}
	else
		{
		p->sort_vector_len = 0;
		p->sort_vector = 0;
		}
	NXReadType( stream, @encode(BOOL), &(p->isOn) );
	NXReadType( stream, @encode(BOOL), &(p->autoSort) );
	if (ver < MISC_TS_VERSION_2)
		{
		BOOL b;
		NXReadType( stream, @encode(BOOL), &b );		// obsolete.
		NXReadType( stream, @encode(BOOL), &b );
		}

	if (p->border == 0)
		p->border = new( NXZoneMalloc( z, sizeof(*(p->border)) ) )
						MiscTableBorder( MISC_ROW_BORDER );
	p->border->read( stream, ver );
	}


//-----------------------------------------------------------------------------
// - read:cells:
// FIXME: What happens when a non-empty non-lazy TableScroll reads the data
// from a lazy, empty one.	Does the memory get freed?
//-----------------------------------------------------------------------------
- (void)read:(int)ver cells:(NXTypedStream*)stream
	{
	NXReadType( stream, @encode(int), &num_cols );
	NXReadType( stream, @encode(int), &num_rows );
	if (lazy)
		{
		max_rows = 0;			// No rows have been allocated.
		max_cells = 0;
		cells = 0;
		}
	else // (!lazy)
		{
		max_rows = num_rows;
		max_cells = max_rows * num_cols;
		if (max_cells == 0)
			cells = 0;
		else
			{
			int const nbytes = max_cells * sizeof(*cells);
			cells = (id*) NXZoneMalloc( [self zone], nbytes );
			id* p = cells;
			for (int r = 0; r < num_rows; r++)
				for (int c = 0; c < num_cols; c++)
					*p++ = NXReadObject( stream );
			}
		}
	}


//-----------------------------------------------------------------------------
// - read:
//-----------------------------------------------------------------------------
- (id)read:(NXTypedStream*)stream
	{
	[super read:stream];
	char const* class_name = [[MiscTableScroll class] name];
	int ver = NXTypedStreamClassVersion( stream, class_name );
	if (((unsigned int)ver) > ((unsigned int)MISC_TS_VERSION))
		[self error:"%s: old library (version %d), can't -read: "
						"new object (version %d).\n",
						class_name, MISC_TS_VERSION, ver ];
	char* s = 0;
	[self read:ver cornerTitle:&s globalInfo:stream];
	[self read:ver border:&colInfo stream:stream];
	[self read:ver border:&rowInfo stream:stream];
	[self read:ver cells:stream];
	[self doInit:ver cornerTitle:s];
	if (s != 0)
		free( s );
	return self;
	}


//-----------------------------------------------------------------------------
// - writeSelector:stream:
//-----------------------------------------------------------------------------
- (void)writeSelector:(SEL)cmd stream:(NXTypedStream*)stream
	{
	char const* str = 0;
	if (cmd != 0)
		str = sel_getName( cmd );
	if (str == 0)						// not an 'else if'
		str = "";
	NXWriteType( stream, @encode(char*), &str );
	}


//-----------------------------------------------------------------------------
// - writeGlobalInfo:
//		NOTE: Cannot archive sort_entry_func and sort_slot_func because they 
//		are function addresses.	 
//-----------------------------------------------------------------------------
- (void)writeGlobalInfo:(NXTypedStream*)stream
	{
	NXWriteType( stream, @encode(int), &tag );
	NXWriteType( stream, @encode(BOOL), &enabled );
	NXWriteType( stream, @encode(BOOL), &lazy );
	NXWriteType( stream, @encode(BOOL), &drawClippedText );
	NXWriteType( stream, @encode(MiscSelectionMode), &mode );
	NXWriteObject( stream, font );
	NXWriteColor( stream, textColor );
	NXWriteColor( stream, backgroundColor );
	NXWriteColor( stream, highlightTextColor );
	NXWriteColor( stream, highlightBackgroundColor );
	NXWriteObjectReference( stream, nextText );
	NXWriteObjectReference( stream, previousText );
	NXWriteObjectReference( stream, delegate );
	NXWriteObjectReference( stream, dataDelegate );
	NXWriteObjectReference( stream, target );
	NXWriteObjectReference( stream, doubleTarget );
	[self writeSelector: action stream: stream];
	[self writeSelector: doubleAction stream: stream];
	char const* const s = [self cornerTitle];
	NXWriteType( stream, @encode(char*), &s );
	}


//-----------------------------------------------------------------------------
// - writeBorder:stream:
//-----------------------------------------------------------------------------
- (void)writeBorder:(MiscBorderInfo*)p stream:(NXTypedStream*)stream
	{
	int const n = p->sort_vector_len;
	NXWriteType( stream, @encode(int), &(n) );
	if (n > 0)
		for (int i = 0;	 i < n;	 i++)
			NXWriteType( stream, @encode(int), &(p->sort_vector[i]) );
	NXWriteType( stream, @encode(BOOL), &(p->isOn) );
	NXWriteType( stream, @encode(BOOL), &(p->autoSort) );
	p->border->write( stream );
	}


//-----------------------------------------------------------------------------
// - writeCells:
//-----------------------------------------------------------------------------
- (void)writeCells:(NXTypedStream*)stream
	{
	NXWriteType( stream, @encode(int), &num_cols );
	NXWriteType( stream, @encode(int), &num_rows );
	if (!lazy)
		for (int r = 0;	 r < num_rows;	r++)
			for (int c = 0;	 c < num_cols;	c++)
				NXWriteObject( stream, [self cellAt:r:c] );
	}


//-----------------------------------------------------------------------------
// - write:
//-----------------------------------------------------------------------------
- (id)write:(NXTypedStream*)stream
	{
	// Ensure that subviews are NOT archived.
	if (colInfo.isOn)	[colInfo.clip removeFromSuperview];
	if (rowInfo.isOn)	[rowInfo.clip removeFromSuperview];
	if (rowInfo.isOn && colInfo.isOn) [cornerView removeFromSuperview];
	[tableView removeFromSuperview];

	[super write:stream];
	[self writeGlobalInfo:stream];
	[self writeBorder:&colInfo stream:stream];
	[self writeBorder:&rowInfo stream:stream];
	[self writeCells:stream];

	// Restore state.
	[self setDocView: tableView];
	if (colInfo.isOn)	[self addSubview:colInfo.clip];
	if (rowInfo.isOn)	[self addSubview:rowInfo.clip];
	if (rowInfo.isOn && colInfo.isOn) [self addSubview:cornerView];
	return self;
	}

@end
