//=============================================================================
//
//	Copyright (C) 1995,1996,1997,1998 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableScrollInspector.M
//
//		Interface Builder inspector for MiscTableScroll.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableScrollInspector.M,v 1.13 98/02/17 14:14:14 zarnuk Exp $
// $Log:		MiscTableScrollInspector.M,v $
//	Revision 1.13  98/02/17	 14:14:14  zarnuk
//	v135: Fixed v130 bug: up/down arrows stopped working.
//	
//	Revision 1.12  97/12/19	 13:57:46  zarnuk
//	v133: Moved the images into the nib directory itself.
//	
//	Revision 1.11  97/12/19	 13:26:24  zarnuk
//	v133: Eliminated constrain-max, constrain-min is now implicit based on
//	presence/absence of auto-size slots.  Eliminated data-sizing.
//	Added alignment control.
//-----------------------------------------------------------------------------
#import "MiscTableScrollInspector.h"
#import "MiscTableBorder.h"
#import <misckit/MiscTableScroll.h>

//=============================================================================
// IB-ONLY METHODS FOR MiscTableScroll
//=============================================================================

@interface MiscTableScroll(IB)
- (char const*)getInspectorClassName;
- (id)editSelf:(id)sender in:(id)owner;
@end

@implementation MiscTableScroll(IB)
- (char const*)getInspectorClassName { return "MiscTableScrollInspector"; }
// Prevent IB from allowing double-click to edit Scroll View
- (id)editSelf:(id)sender in:(id)owner { return 0; }
@end

//=============================================================================
// CONVENIENCE CATAGORIES
//=============================================================================

@interface Matrix(MiscPopUpList)
- (int)popUpTag;
@end

@implementation Matrix(MiscPopUpList)
- (int)popUpTag { return [[self selectedCell] tag]; }
@end


@interface Button(MiscPopUpList)
- (int)popUpTag;
- (void)setPopUpTag:(int)new_tag;
@end

@implementation Button(MiscPopUpList)
- (int)popUpTag
	{
	id popUp = [self target];
	return [[[popUp itemList] selectedCell] tag];
	}

- (void)setPopUpTag:(int)new_tag
	{
	id popUp = [self target];
	Matrix* matrix = [popUp itemList];
	MenuCell* m_cell = [matrix findCellWithTag: new_tag];
	[matrix selectCell: m_cell];
	[self setTitle: [m_cell title]];
	}
@end



//=============================================================================
// IMPLEMENTATION
//=============================================================================
@interface MiscTableScrollInspector(Forward)
- (id)doSlot:(id)sender;
- (id)doTitleMode:(id)sender;
@end

@implementation MiscTableScrollInspector

- (BOOL) wantsButtons { return NO; }
- (id)ok:(id)sender { return [super ok:sender]; }


//-----------------------------------------------------------------------------
// init
//-----------------------------------------------------------------------------
- (id)init
	{
	[super init];
	char buf[ MAXPATHLEN + 1 ];
	id bundle = [NXBundle bundleForClass: [MiscTableScrollInspector class]];
	[bundle getPath: buf forResource: [[self class] name] ofType: "nib"];
	[NXApp loadNibFile: buf owner: self withNames: NO fromZone: [self zone]];
	border_type = MISC_COL_BORDER;
	dirty = NO;
	return self;
	}


//-----------------------------------------------------------------------------
// setTitleControls
//-----------------------------------------------------------------------------
- (void)setTitleControls
	{
	if ([object borderTitlesOn:border_type])
		{
		[titlesSwitch setState:1];
		[[sizableSwitch setEnabled: YES] setState:border->isSizeable()];
		[[draggableSwitch setEnabled:YES] setState:border->isDraggable()];
		[[titleModePopUp setEnabled:YES]
				setPopUpTag:(int)[object borderTitleMode:border_type]];
		}
	else
		{
		[titlesSwitch setState:0];
		[[sizableSwitch setEnabled:NO] setState:0];
		[[draggableSwitch setEnabled:NO] setState:0];
		[[titleModePopUp setEnabled:NO]
				setPopUpTag:(int)[object borderTitleMode:border_type]];
		}
	[[uniformSizeField setEnabled: YES]
			setIntValue: int([object uniformSizeSlots:border_type])];
	}


//-----------------------------------------------------------------------------
// fillScroll
//-----------------------------------------------------------------------------
- (void)fillScroll
	{
	[window disableFlushWindow];
	num_slots = [object numSlots:border_type];
	if (num_slots == 0)
		{
		[slotScroll clearSelectedCell];
		[slotScroll renewRows:0 cols:0]; // Gets rid of horiz scroller knob
		}
	else
		{
		[slotScroll renewRows:num_slots cols:1];
		for (int i = 0;	 i < num_slots;	 i++)
			{
			int const pslot = [object border:border_type slotAtPosition:i];
			[[slotScroll cellAt:i:0] setTitle: [object border:border_type
						slotTitle:pslot]];
			}
		}
	[slotScroll sizeToCells];
	[slotScroll display];

	if (num_slots > 0)
		{
		if (slot < 0) slot = 0;
		[slotScroll selectCellAt:slot:0];
		[slotScroll scrollCellToVisible:slot:0];
		}
	else
		{
		slot = -1;
		}
	[self doSlot:self];
	[[window reenableFlushWindow] flushWindow];
	}


//-----------------------------------------------------------------------------
// doEnabled:
//-----------------------------------------------------------------------------
- (id)doEnabled:(id)sender
	{
	[object setEnabled: [enabledSwitch state]];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doLazy:
//-----------------------------------------------------------------------------
- (id)doLazy:(id)sender
	{
	[object setLazy: [lazySwitch state]];
	[self ok:sender];
	[self revert:self];
	return self;
	}


//-----------------------------------------------------------------------------
// doMode:
//-----------------------------------------------------------------------------
- (id)doMode:(id)sender
	{
	[object setSelectionMode: MiscSelectionMode([modePopUp popUpTag])];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doBorder:
//-----------------------------------------------------------------------------
- (id)doBorder:(id)sender
	{
	MiscBorderType bt = MiscBorderType([borderPopUp popUpTag]);
	if (bt != border_type)
		{
		border_type = bt;
		[self revert:sender];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// doTitles:
//-----------------------------------------------------------------------------
- (id)doTitles:(id)sender
	{
	BOOL const isOn = [titlesSwitch state];
	[object border:border_type setTitlesOn:isOn];
	[self setTitleControls];
	[self doSlot:self];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doSizable:
//-----------------------------------------------------------------------------
- (id)doSizable:(id)sender
	{
	[object border:border_type setSizeableSlots:[sizableSwitch state]];
	[self doSlot:0];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doDraggable:
//-----------------------------------------------------------------------------
- (id)doDraggable:(id)sender
	{
	[object border:border_type setDraggableSlots:[draggableSwitch state]];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doAutoSort:
//-----------------------------------------------------------------------------
- (id)doAutoSort:(id)sender
	{
	[object border:border_type setAutoSortSlots:[autoSortSwitch state]];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doTitleMode:
//-----------------------------------------------------------------------------
- (id)doTitleMode:(id)sender
	{
	[object border:border_type
		setTitleMode: (MiscTableTitleMode) [titleModePopUp popUpTag]];
	[self fillScroll];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doDelete:
//-----------------------------------------------------------------------------
- (id)doDelete:(id)sender
	{
	[object border:border_type
				deleteSlotAt:[object border:border_type slotAtPosition:slot]];
	[object tile];
	[object display];
	slot--;
	num_slots--;
	if (slot < 0 && num_slots > 0)
		slot = 0;
	[self fillScroll];
	[self ok:sender];
	return self;
	}


//-----------------------------------------------------------------------------
// doNew:
//-----------------------------------------------------------------------------
- (id)doNew:(id)sender
	{
	[object addSlot:border_type];
	[object sizeToCells];
	slot = num_slots;
	[self fillScroll];
	[self ok:sender];
	return self;
	}


//-----------------------------------------------------------------------------
// doUserSizeable:
//-----------------------------------------------------------------------------
- (id)doUserSizeable:(id)sender
	{
	[object border:border_type
				setSlot:[object border:border_type slotAtPosition:slot]
				sizeable:[userSizeableSwitch state]];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doAlign:
//-----------------------------------------------------------------------------
- (id)doAlign:(id)sender
	{
	int const pslot = [object border:border_type slotAtPosition:slot];
	if ([object border:border_type slotCellType:pslot] == MISC_TABLE_CELL_TEXT)
		{
		id const proto = [object border:border_type slotCellPrototype:pslot];
		id const cell = [alignMatrix selectedCell];
		[proto setAlignment:[cell tag]];
		}
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doAutosize:
//-----------------------------------------------------------------------------
- (id)doAutosize:(id)sender
	{
	[object border:border_type
				setSlot:[object border:border_type slotAtPosition:slot]
				autosize:[autosizeSwitch state]];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doCellClass:
//-----------------------------------------------------------------------------
- (id)doCellClass:(id)sender
	{
	int const pslot = [object border:border_type slotAtPosition:slot];
	int const old_val = (int)[object border:border_type slotCellType:pslot];
	int const new_val = [cellClassPopUp popUpTag];

	if (old_val != new_val)
		{
		[object border:border_type setSlot:pslot
				cellType: (MiscTableCellStyle) new_val];
		if (new_val == MISC_TABLE_CELL_TEXT)
			{
			[alignTitle setTextColor:NX_COLORBLACK];
			[alignMatrix setEnabled:YES];
			id const cell = [object border:border_type slotCellPrototype:pslot];
			[alignMatrix selectCellWithTag:[cell alignment]];
			}
		else if (old_val == MISC_TABLE_CELL_TEXT)
			{
			[alignTitle setTextColor:NX_COLORDKGRAY];
			[alignMatrix setEnabled:NO];
			}
		}

	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doSortType:
//-----------------------------------------------------------------------------
- (id)doSortType:(id)sender
	{
	[object border:border_type
		setSlot:[object border:border_type slotAtPosition:slot]
		sortType: (MiscSortType) [sortTypePopUp popUpTag]];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doSortDirection:
//-----------------------------------------------------------------------------
- (id)doSortDirection:(id)sender
	{
	[object border:border_type
		setSlot:[object border:border_type slotAtPosition:slot]
		sortDirection: (MiscSortDirection) [sortDirectionPopUp popUpTag]];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doSlot:
//-----------------------------------------------------------------------------
- (id)doSlot:(id)sender
	{
	slot = [slotScroll selectedRow];
	if (slot >= 0)
		{
		BOOL const titlesOn = [object borderTitlesOn:border_type];
		BOOL const notUniform = ([object uniformSizeSlots:border_type] == 0);
		int pslot = [object border:border_type slotAtPosition:slot];
		[deleteButton setEnabled:YES];
		[upButton setEnabled:(slot > 0)];
		[downButton setEnabled:(slot < num_slots - 1)];
		int const ctype = (int)[object border:border_type slotCellType:pslot];
		[[cellClassPopUp setEnabled:YES] setPopUpTag:ctype];
		[[sortTypePopUp setEnabled:YES] setPopUpTag:
				(int)[object border:border_type slotSortType:pslot]];
		[[sortDirectionPopUp setEnabled:YES] setPopUpTag:
				(int)[object border:border_type slotSortDirection:pslot]];
		if (ctype == MISC_TABLE_CELL_TEXT)
			{
			[alignTitle setTextColor:NX_COLORBLACK];
			[alignMatrix setEnabled:YES];
			id const cell = [object border:border_type slotCellPrototype:pslot];
			[alignMatrix selectCellWithTag:[cell alignment]];
			}
		else
			{
			[alignTitle setTextColor:NX_COLORDKGRAY];
			[alignMatrix setEnabled:NO];
			}
		[[autosizeSwitch setEnabled:notUniform]
				setState:[object border:border_type slotIsAutosize:pslot]];
		BOOL const enabled = titlesOn && notUniform &&
				[object sizeableSlots:border_type];
		[[userSizeableSwitch setEnabled: enabled] setState: enabled &&
				[object border:border_type slotIsSizeable:pslot]];
		[[sizeField setEnabled:notUniform] setIntValue:
				(int)[object border:border_type slotSize:pslot]];
		[[sizeMinField setEnabled:notUniform] setIntValue:
				(int)[object border:border_type slotMinSize:pslot]];
		[[sizeMaxField setEnabled:notUniform] setIntValue:
				(int)[object border:border_type slotMaxSize:pslot]];
		[[titleField setEnabled: titlesOn &&
				([object borderTitleMode:border_type] == MISC_CUSTOM_TITLE)]
				setStringValue: [object border:border_type slotTitle:pslot]];
		[titleField selectText: self];
		}
	else
		{
		[deleteButton setEnabled:NO];
		[upButton setEnabled:NO];
		[downButton setEnabled:NO];
		[[cellClassPopUp setEnabled:NO] setPopUpTag:0];
		[[sortTypePopUp setEnabled:NO] setPopUpTag:0];
		[[sortDirectionPopUp setEnabled:NO] setPopUpTag:0];
		[alignTitle setTextColor:NX_COLORDKGRAY];
		[alignMatrix setEnabled:NO];
		[alignMatrix selectCellWithTag:0];
		[[autosizeSwitch setEnabled:NO] setState:0];
		[[userSizeableSwitch setEnabled:NO] setState:0];
		[[sizeField setEnabled:NO] setStringValue:""];
		[[sizeMinField setEnabled:NO] setStringValue:""];
		[[sizeMaxField setEnabled:NO] setStringValue:""];
		[[titleField setEnabled:NO] setStringValue:""];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// swapSlots::
//-----------------------------------------------------------------------------
- (void)swapSlots:(int)from_slot :(int)to_slot
	{
	border->swapSlots( from_slot, to_slot );
	[object display];
	slot = to_slot;
	[self fillScroll];
	[self ok:self];
	}


//-----------------------------------------------------------------------------
// doUp:
//-----------------------------------------------------------------------------
- (id)doUp:(id)sender
	{
	if (slot > 0)
		[self swapSlots:slot :slot-1];
	return self;
	}


//-----------------------------------------------------------------------------
// doDown:
//-----------------------------------------------------------------------------
- (id)doDown:(id)sender
	{
	if (slot < num_slots - 1)
		[self swapSlots:slot :slot+1];
	return self;
	}


//-----------------------------------------------------------------------------
// get_pslot
//		Get the physical slot-id of the current slot.
//-----------------------------------------------------------------------------
- (int) get_pslot
	{
	return [object border:border_type slotAtPosition:slot];
	}


//-----------------------------------------------------------------------------
// textDidChange:
//-----------------------------------------------------------------------------
- (id)textDidChange:(id)sender
	{
	dirty = YES;
	return [super textDidChange:sender];
	}


//-----------------------------------------------------------------------------
// textDidEnd:endChar:
//-----------------------------------------------------------------------------
- (id)textDidEnd:(id)sender endChar:(unsigned short)whyEnd
	{
	if (dirty)
		{
		dirty = NO;
		id field = 0;
		if ([sender respondsTo:@selector(delegate)])
			field = [sender delegate];
		if (field == (id)titleField)
			{
			int const pslot = [self get_pslot];
			[object border:border_type setSlot:pslot
											title:[titleField stringValue]];
			[[slotScroll cellAt:slot:0] setTitle:[titleField stringValue]];
			[slotScroll drawCellAt:slot:0];
			}
		else if (field == (id)uniformSizeField)
			{
			int old_size = (int) [object uniformSizeSlots:border_type];
			int new_size = [uniformSizeField intValue];
			if (new_size != 0)
				{
				if (new_size < MISC_MIN_PIXELS_SIZE)
					new_size = MISC_MIN_PIXELS_SIZE;
				if (new_size > MISC_MAX_PIXELS_SIZE)
					new_size = MISC_MAX_PIXELS_SIZE;
				[uniformSizeField setIntValue:new_size];
				}
			if (old_size != new_size)
				{
				[object border:border_type
						setUniformSizeSlots:(NXCoord)new_size];
				[self ok:sender];
				[self revert:0];
				}
			}
		else if (field == (id)sizeField)
			{
			int const pslot = [self get_pslot];
			int i = [sizeField intValue];
			int const imin = (int)[object border:border_type slotMinSize:pslot];
			int const imax = (int)[object border:border_type slotMaxSize:pslot];

			if (i < imin || i > imax)
				{
				if (i < imin) i = imin;
				if (i > imax) i = imax;
				[sizeField setIntValue:i];
				}
			[object border:border_type setSlot:pslot size:(NXCoord)i];
			}
		else if (field == (id)sizeMinField)
			{
			int const pslot = [self get_pslot];
			int const i = (int)[object border:border_type slotSize:pslot];
			int imin = [sizeMinField intValue];
			int const imax = (int)[object border:border_type slotMaxSize:pslot];

			if (imin < MISC_MIN_PIXELS_SIZE)
				imin = MISC_MIN_PIXELS_SIZE;
			else if (imin > imax)
				imin = imax;
			if (imin > i)		// Not an 'else if'
				imin = i;

			[sizeMinField setIntValue:imin];
			[object border:border_type setSlot:pslot minSize:(NXCoord)imin];
			}
		else if (field == (id)sizeMaxField)
			{
			int const pslot = [self get_pslot];
			int const i = (int)[object border:border_type slotSize:pslot];
			int const imin = (int)[object border:border_type slotMinSize:pslot];
			int imax = [sizeMaxField intValue];

			if (imax > MISC_MAX_PIXELS_SIZE)
				imax = MISC_MAX_PIXELS_SIZE;
			else if (imax < imin)
				imax = imin;
			if (imax < i)		// Not an 'else if'
				imax = i;

			[sizeMaxField setIntValue:imax];
			[object border:border_type setSlot:pslot maxSize:(NXCoord)imax];
			}
		[self ok: sender];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// doColorText
//-----------------------------------------------------------------------------
- (id)doColorText:(id)sender
	{
	[object setTextColor: [sender color]];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doColorBack
//-----------------------------------------------------------------------------
- (id)doColorBack:(id)sender
	{
	[object setBackgroundColor: [sender color]];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doColorTextH
//-----------------------------------------------------------------------------
- (id)doColorTextH:(id)sender
	{
	[object setHighlightTextColor: [sender color]];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// doColorBackH
//-----------------------------------------------------------------------------
- (id)doColorBackH:(id)sender
	{
	[object setHighlightBackgroundColor: [sender color]];
	return [self ok:sender];
	}


//-----------------------------------------------------------------------------
// revert:
//-----------------------------------------------------------------------------
- (id)revert:(id)sender
	{
	dirty = NO;
	[super revert: sender];
	[window disableFlushWindow];

	border = [object border:border_type];
	[autoSortSwitch setState:[object autoSortSlots:border_type]];
	[self setTitleControls];

	[enabledSwitch setState: [object isEnabled]];
	[lazySwitch setState: [object isLazy]];
	[modePopUp setPopUpTag:(int) [object selectionMode]];

	[colorText	setColor: [object textColor]];
	[colorTextH setColor: [object highlightTextColor]];
	[colorBack	setColor: [object backgroundColor]];
	[colorBackH setColor: [object highlightBackgroundColor]];

	slot = -1;
	[self fillScroll];

	id fr = [window firstResponder];
	if (fr == 0 || fr == window)
		[titleField selectText: self];
	[[window reenableFlushWindow] flushWindow];
	return self;
	}

@end
