#ifndef __MiscTableScrollPrivate_h
#define __MiscTableScrollPrivate_h
//=============================================================================
//
//		Copyright (C) 1996,1997 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableScrollPrivate.h
//
//		Methods used within the Table Scroll palette but which are not
//		exported with the *public* interface.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableScrollPrivate.h,v 1.8 97/12/19 01:11:28 zarnuk Exp $
// $Log:		MiscTableScrollPrivate.h,v $
// Revision 1.8	 97/12/19  01:11:28	 zarnuk
// v133: Broke off IO category of TableScroll.	Migrated some version
// related declarations here.  Added access to the cornerView.
// 
// Revision 1.7	 97/02/05  07:57:56	 sunshine
// v122: Added -setTracking: to fix lazy-mode mouse-tracking display problem.
//-----------------------------------------------------------------------------
#import <misckit/MiscTableScroll.h>
#import "MiscDelegateFlags.h"

// There are no structural differences between versions 0 and 1.  However,
// the range of the MiscSortType values increased from 0..8 to 0..12
// There are significant structural differences between versions 1 and 2.
#define MISC_TS_VERSION_0		0
#define MISC_TS_VERSION_1		1
#define MISC_TS_VERSION_2		2
#define MISC_TS_VERSION			MISC_TS_VERSION_2

@interface MiscTableScroll (PrivateInternal)

- (id)responsibleDelegate:(MiscDelegateFlags::Selector)cmd;

// Private: border -> scroll
- (char const*)border:(MiscBorderType)b getDelegateSlotTitle:(int)slot;
- (id)border:(MiscBorderType)b getDelegateSlotPrototype:(int)s;

// TableScroll(IncrementalSearch) -> TableScroll(Keyboard)
- (void)keyboardSelect:(NXEvent const*)p;

- (MiscBorderView*)rowTitles;
- (MiscBorderView*)colTitles;
- (MiscCornerView*)cornerView;

- (void)setTracking:(BOOL)flag;
- (void)setClickedRow:(MiscCoord_P)r col:(MiscCoord_P)c;
- (void)clearClicked;

// Reset stale-old-selection in Table & Border views.
- (void)resetSelection;

// TableScroll(IO) -> TableScroll
- (void)doInit:(int)ver cornerTitle:(char const*)s;

@end

#endif // __MiscTableScrollPrivate_h
