//=============================================================================
//
//		Copyright (C) 1995-1997 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableScrollSel.M
//
//		Selection management methods for MiscTableScroll.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableScrollSel.M,v 1.1 97/11/14 20:41:24 zarnuk Exp $
// $Log:		MiscTableScrollSel.M,v $
//	Revision 1.1  97/11/14	20:41:24  zarnuk
//	v130: selection, keyboard and mouse routines for MiscTableScroll.
//	
//-----------------------------------------------------------------------------
#import <misckit/MiscTableScroll.h>
#import "MiscColView.h"
#import "MiscRowView.h"
#import "MiscTableView.h"
#import "MiscTableBorder.h"

@implementation MiscTableScroll(Selection)
//-----------------------------------------------------------------------------
// Next / Previous Text
//-----------------------------------------------------------------------------
- (id)nextText			{ return nextText; }
- (id)previousText		{ return previousText; }
- (id)setNextText:(id)obj
	{
	nextText = obj;
	if (obj && [obj respondsTo: @selector(setPreviousText:)])
		[obj setPreviousText: self];
	return self;
	}
- (id)setPreviousText:(id)obj	{ previousText = obj; return self; }
- (BOOL)isSelectable			{ return num_cols > 0 && num_rows > 0; }
- (BOOL)isEnabled				{ return enabled; }
- (id)setEnabled:(BOOL)flag		{ enabled = flag; return self; }

- (id)selectText:(id)sender
	{
	if ([window firstResponder] != (id)tableView)
		[window makeFirstResponder:tableView];
	return self;
	}


//-----------------------------------------------------------------------------
// Selection
//-----------------------------------------------------------------------------
- (void)selectionChanged
	{
	[window disableFlushWindow];
	[rowInfo.view selectionChanged];
	[colInfo.view selectionChanged];
	[tableView selectionChanged];
	[[window reenableFlushWindow] flushWindow];
	}

- (void)resetSelection
	{
	[tableView resetSelection];
	[rowInfo.view resetSelection];
	[colInfo.view resetSelection];
	}

- (void)trackBy:(MiscBorderType)b		{ [tableView trackBy:b]; }
- (MiscBorderType)trackingBy			{ return [tableView trackingBy]; }
- (BOOL)tracking						{ return tracking; }
- (void)setTracking:(BOOL)flag			{ tracking = flag; }

- (MiscSelectionMode)selectionMode		{ return mode; }
- (void)setSelectionMode:(MiscSelectionMode)x
	{
	if (x != mode)
		{
		[self clearSelection];
		mode = x;
		[colInfo.view setSelectionMode: mode];
		[rowInfo.view setSelectionMode: mode];
		[tableView setSelectionMode: mode];
		}
	}

- (id)selectedCell
	{
	MiscCoord_P const col = [self selectedCol];
	MiscCoord_P const row = [self selectedRow];
	return (col >= 0 && row >= 0) ? [self cellAt:row:col] : 0;
	}

- (BOOL)borderHasSelection:(MiscBorderType)b
		{ return info[b]->border->hasSelection(); }
- (BOOL)hasRowSelection { return [self borderHasSelection:MISC_ROW_BORDER]; }
- (BOOL)hasColSelection { return [self borderHasSelection:MISC_COL_BORDER]; }

- (BOOL)borderHasMultipleSelection:(MiscBorderType)b
		{ return info[b]->border->hasMultipleSelection(); }
- (BOOL)hasMultipleRowSelection
		{ return [self borderHasMultipleSelection: MISC_ROW_BORDER]; }
- (BOOL)hasMultipleColSelection
		{ return [self borderHasMultipleSelection: MISC_COL_BORDER]; }

- (unsigned int)borderNumSelectedSlots:(MiscBorderType)b
		{ return info[b]->border->numSelected(); }
- (unsigned int)numSelectedRows
		{ return [self borderNumSelectedSlots: MISC_ROW_BORDER]; }
- (unsigned int)numSelectedCols
		{ return [self borderNumSelectedSlots: MISC_COL_BORDER]; }

- (BOOL)border:(MiscBorderType)b slotIsSelected:(MiscCoord_P)slot
		{
		MiscTableBorder const* bp = info[b]->border;
		return bp->goodPos( slot ) && bp->isSelected_P( slot );
		}
- (BOOL)rowIsSelected:(MiscCoord_P)row
		{ return [self border:MISC_ROW_BORDER slotIsSelected:row]; }
- (BOOL)colIsSelected:(MiscCoord_P)col
		{ return [self border:MISC_COL_BORDER slotIsSelected:col]; }
- (BOOL)cellIsSelected:(MiscCoord_P)r :(MiscCoord_P)c
		{ return ([self rowIsSelected:r] || [self colIsSelected:c]); }

- (MiscCoord_P)borderSelectedSlot:(MiscBorderType)b
		{ return info[b]->border->selectedSlot_P(); }
- (MiscCoord_P)selectedRow
		{ return [self borderSelectedSlot:MISC_ROW_BORDER]; }
- (MiscCoord_P)selectedCol
		{ return [self borderSelectedSlot:MISC_COL_BORDER]; }

- (void)border:(MiscBorderType)b selectedTags:(MiscIntList*)tags
		{ info[b]->border->selectedTags( tags ); }
- (void)selectedRowTags:(MiscIntList*)tags
		{ [self border:MISC_ROW_BORDER selectedTags:tags]; }
- (void)selectedColTags:(MiscIntList*)tags
		{ [self border:MISC_COL_BORDER selectedTags:tags]; }

- (void)border:(MiscBorderType)b selectedSlots:(MiscIntList*)slots
		{ info[b]->border->selectedSlots( slots ); }
- (void)selectedRows:(MiscIntList*)rows
		{ [self border:MISC_ROW_BORDER selectedSlots:rows]; }
- (void)selectedCols:(MiscIntList*)cols
		{ [self border:MISC_COL_BORDER selectedSlots:cols]; }

- (void)border:(MiscBorderType)b selectSlot:(MiscCoord_P)p_slot
		byExtension:(BOOL)extend
		{
		BOOL changed = NO;
		MiscTableBorder* const ob = info[ MISC_OTHER_BORDER(b) ]->border;
		if (ob->hasSelection())
			{
			ob->selectNone();
			changed = YES;
			}
		MiscTableBorder* const bp = info[b]->border;
		if (!extend && bp->numSelected() > 0)
			{
			bp->selectNone();
			changed = YES;
			}
		if (bp->goodPos( p_slot ))
			{
			MiscCoord_V const v_slot = bp->physicalToVisual( p_slot );
			if (!bp->isSelected( v_slot ))
				{
				if (mode == MISC_RADIO_MODE)
					bp->selectOne( v_slot );
				else
					bp->select( v_slot );
				changed = YES;
				}
			}
		if (changed)
			[self selectionChanged];
		}
- (void)border:(MiscBorderType)b selectSlot:(MiscCoord_P)p_slot
		{ [self border:b selectSlot:p_slot byExtension:NO]; }
- (void)selectRow:(MiscCoord_P)row byExtension:(BOOL)extend
		{ [self border:MISC_ROW_BORDER selectSlot:row byExtension:extend]; }
- (void)selectRow:(MiscCoord_P)row
		{ [self selectRow:row byExtension:NO]; }
- (void)selectCol:(MiscCoord_P)col byExtension:(BOOL)extend
		{ [self border:MISC_COL_BORDER selectSlot:col byExtension:extend]; }
- (void)selectCol:(MiscCoord_P)col
		{ [self selectCol:col byExtension:NO]; }

- (void)border:(MiscBorderType)b selectTags:(MiscIntList*)tags
		byExtension:(BOOL)extend
		{
		info[b]->border->selectTags( tags, extend );
		[self selectionChanged];
		}
- (void)border:(MiscBorderType)b selectTags:(MiscIntList*)tags
		{ [self border:b selectTags:tags byExtension:NO]; }
- (void)selectRowTags:(MiscIntList*)tags byExtension:(BOOL)extend
		{ [self border:MISC_ROW_BORDER selectTags:tags byExtension:extend]; }
- (void)selectRowTags:(MiscIntList*)tags
		{ [self selectRowTags:tags byExtension:NO]; }
- (void)selectColTags:(MiscIntList*)tags byExtension:(BOOL)extend
		{ [self border:MISC_COL_BORDER selectTags:tags byExtension:extend]; }
- (void)selectColTags:(MiscIntList*)tags
		{ [self selectColTags:tags byExtension:NO]; }

- (void)border:(MiscBorderType)b selectSlots:(MiscIntList*)slots
		byExtension:(BOOL)extend
		{
		info[b]->border->selectSlots( slots, extend );
		[self selectionChanged];
		}
- (void)border:(MiscBorderType)b selectSlots:(MiscIntList*)slots
		{ [self border:b selectSlots:slots byExtension:NO]; }
- (void)selectRows:(MiscIntList*)rows byExtension:(BOOL)extend
		{ [self border:MISC_ROW_BORDER selectSlots:rows byExtension:extend]; }
- (void)selectRows:(MiscIntList*)rows
		{ [self selectRows:rows byExtension:NO]; }
- (void)selectCols:(MiscIntList*)cols byExtension:(BOOL)extend
		{ [self border:MISC_COL_BORDER selectSlots:cols byExtension:extend]; }
- (void)selectCols:(MiscIntList*)cols
		{ [self selectCols:cols byExtension:NO]; }

- (void)borderSelectAll:(MiscBorderType)b
		{
		if (mode == MISC_LIST_MODE || mode == MISC_HIGHLIGHT_MODE)
			{
			info[b]->border->selectAll();
			[self selectionChanged];
			}
		}
- (void)selectAllRows { [self borderSelectAll: MISC_ROW_BORDER]; }
- (void)selectAllCols { [self borderSelectAll: MISC_COL_BORDER]; }
- (id)selectAll:(id)sender
		{
		[self selectAllRows];
		[self sendActionIfEnabled];
		return self;
		}

- (void)border:(MiscBorderType)b deselectSlot:(MiscCoord_P)p_slot
		{
		MiscTableBorder* const bp = info[b]->border;
		if (bp->goodPos( p_slot ))
			{
			MiscCoord_V const v_slot = bp->physicalToVisual( p_slot );
			if (bp->isSelected( v_slot ))
				{
				bp->unselect( v_slot );
				[self selectionChanged];
				}
			}
		}
- (void)deselectRow:(MiscCoord_P)row
		{ [self border:MISC_ROW_BORDER deselectSlot:row]; }
- (void)deselectCol:(MiscCoord_P)col
		{ [self border:MISC_COL_BORDER deselectSlot:col]; }

- (void)border:(MiscBorderType)b deselectTags:(MiscIntList*)tags
		{
		info[b]->border->unselectTags( tags );
		[self selectionChanged];
		}
- (void)deselectRowTags:(MiscIntList*)tags
		{ [self border:MISC_ROW_BORDER deselectTags:tags]; }
- (void)deselectColTags:(MiscIntList*)tags
		{ [self border:MISC_COL_BORDER deselectTags:tags]; }

- (void)border:(MiscBorderType)b deselectSlots:(MiscIntList*)slots
		{
		info[b]->border->unselectSlots( slots );
		[self selectionChanged];
		}
- (void)deselectRows:(MiscIntList*)rows
		{ [self border:MISC_ROW_BORDER deselectSlots:rows]; }
- (void)deselectCols:(MiscIntList*)cols
		{ [self border:MISC_COL_BORDER deselectSlots:cols]; }

- (void)borderClearSelection:(MiscBorderType)b
		{
		info[b]->border->selectNone();
		[self selectionChanged];
		}
- (void)clearRowSelection { [self borderClearSelection: MISC_ROW_BORDER]; }
- (void)clearColSelection { [self borderClearSelection: MISC_COL_BORDER]; }
- (void)clearSelection { [self clearRowSelection]; [self clearColSelection]; }
- (id)deselectAll:(id)sender
		{
		[self clearSelection];
		[self sendActionIfEnabled];
		return self;
		}


//-----------------------------------------------------------------------------
// Clicked slot stuff
//-----------------------------------------------------------------------------
- (MiscCoord_P)borderClickedSlot:(MiscBorderType)b
		{ return info[b]->border->clickedSlot_P(); }
- (MiscCoord_P)clickedRow { return [self borderClickedSlot:MISC_ROW_BORDER]; }
- (MiscCoord_P)clickedCol { return [self borderClickedSlot:MISC_COL_BORDER]; }
- (id)clickedCell
		{
		MiscCoord_P const r = [self clickedRow];
		MiscCoord_P const c = [self clickedCol];
		return (r >= 0 && c >= 0 ? [self cellAt:r:c] : 0);
		}

- (void)border:(MiscBorderType)b setClickedSlot:(MiscCoord_P)n
		{ info[b]->border->setClickedSlot_P(n); }
- (void)setClickedRow:(MiscCoord_P)r col:(MiscCoord_P)c
		{
		[self border:MISC_ROW_BORDER setClickedSlot:r];
		[self border:MISC_COL_BORDER setClickedSlot:c];
		}

- (void)borderClearClickedSlot:(MiscBorderType)b
		{ info[b]->border->clearClickedSlot(); }
- (void)clearClicked
		{
		[self borderClearClickedSlot:MISC_ROW_BORDER];
		[self borderClearClickedSlot:MISC_COL_BORDER];
		}


//-----------------------------------------------------------------------------
// Keyboard cursor stuff
//-----------------------------------------------------------------------------
- (void)reflectCursor { [tableView reflectCursor]; }
- (MiscCoord_P)borderCursor: (MiscBorderType)b
		{ return info[b]->border->getCursor_P(); }
- (MiscCoord_P)cursorRow { return [self borderCursor: MISC_ROW_BORDER]; }
- (MiscCoord_P)cursorCol { return [self borderCursor: MISC_COL_BORDER]; }

- (void) border:(MiscBorderType)b setCursor:(MiscCoord_P)slot
		{ info[b]->border->setCursor_P( slot ); [self reflectCursor]; }
- (void)setCursorRow:(MiscCoord_P)row
		{ [self border: MISC_ROW_BORDER setCursor: row]; }
- (void)setCursorCol:(MiscCoord_P)col
		{ [self border: MISC_COL_BORDER setCursor: col]; }

- (void)borderClearCursor:(MiscBorderType)b
		{ info[b]->border->clearCursor(); [self reflectCursor]; }
- (void)clearCursorRow { [self borderClearCursor: MISC_ROW_BORDER]; }
- (void)clearCursorCol { [self borderClearCursor: MISC_COL_BORDER]; }
- (void)clearCursor { [self clearCursorRow]; [self clearCursorCol]; }

- (BOOL)borderHasValidCursor:(MiscBorderType)b
		{ return info[b]->border->hasValidCursor(); }
- (BOOL)hasValidCursorRow
		{ return [self borderHasValidCursor: MISC_ROW_BORDER]; }
- (BOOL)hasValidCursorCol
		{ return [self borderHasValidCursor: MISC_COL_BORDER]; }

- (BOOL)isCursorEnabled { return [tableView isCursorEnabled]; }
- (void)disableCursor
	{
	if ([self isCursorEnabled] && [tableView shouldDrawCursor])
		[tableView eraseCursor];
	[tableView disableCursor];
	}

- (void)reenableCursor
	{
	[tableView reenableCursor];
	if ([self isCursorEnabled] && [tableView shouldDrawCursor])
		[tableView drawCursor];
	}

@end
